import logging
from dataclasses import dataclass
from typing import Any, Dict, Generic, Optional, Type, TypeVar, cast

import pydantic
from pydantic import root_validator
from pydantic.fields import Field
from pydantic.generics import GenericModel

from datahub.configuration.common import (
    ConfigModel,
    ConfigurationError,
    DynamicTypedConfig,
    LineageConfig,
)
from datahub.configuration.time_window_config import BaseTimeWindowConfig
from datahub.ingestion.api.common import PipelineContext
from datahub.ingestion.api.ingestion_job_checkpointing_provider_base import (
    IngestionCheckpointingProviderBase,
    JobId,
)
from datahub.ingestion.api.source import Source, SourceReport
from datahub.ingestion.source.state.checkpoint import Checkpoint, StateType
from datahub.ingestion.source.state.use_case_handler import (
    StatefulIngestionUsecaseHandlerBase,
)
from datahub.ingestion.source.state_provider.state_provider_registry import (
    ingestion_checkpoint_provider_registry,
)
from datahub.metadata.schema_classes import (
    DatahubIngestionCheckpointClass,
    DatahubIngestionRunSummaryClass,
)

logger: logging.Logger = logging.getLogger(__name__)


class DynamicTypedStateProviderConfig(DynamicTypedConfig):
    # Respecifying the base-class just to override field level docs

    type: str = Field(
        description="The type of the state provider to use. For DataHub use `datahub`",
    )
    # This config type is declared Optional[Any] here. The eventual parser for the
    # specified type is responsible for further validation.
    config: Optional[Any] = Field(
        default=None,
        description="The configuration required for initializing the state provider. Default: The datahub_api config if set at pipeline level. Otherwise, the default DatahubClientConfig. See the defaults (https://github.com/datahub-project/datahub/blob/master/metadata-ingestion/src/datahub/ingestion/graph/client.py#L19).",
    )


class StatefulIngestionConfig(ConfigModel):
    """
    Basic Stateful Ingestion Specific Configuration for any source.
    """

    enabled: bool = Field(
        default=False,
        description="The type of the ingestion state provider registered with datahub.",
    )
    max_checkpoint_state_size: pydantic.PositiveInt = Field(
        default=2**24,  # 16 MB
        description="The maximum size of the checkpoint state in bytes. Default is 16MB",
        hidden_from_docs=True,
    )
    state_provider: Optional[DynamicTypedStateProviderConfig] = Field(
        default=None,
        description="The ingestion state provider configuration.",
        hidden_from_docs=True,
    )
    ignore_old_state: bool = Field(
        default=False,
        description="If set to True, ignores the previous checkpoint state.",
    )
    ignore_new_state: bool = Field(
        default=False,
        description="If set to True, ignores the current checkpoint state.",
    )

    @pydantic.root_validator()
    def validate_config(cls, values: Dict[str, Any]) -> Dict[str, Any]:
        if values.get("enabled"):
            if values.get("state_provider") is None:
                values["state_provider"] = DynamicTypedStateProviderConfig(
                    type="datahub", config=None
                )
        return values


CustomConfig = TypeVar("CustomConfig", bound=StatefulIngestionConfig)


class StatefulIngestionConfigBase(GenericModel, Generic[CustomConfig]):
    """
    Base configuration class for stateful ingestion for source configs to inherit from.
    """

    stateful_ingestion: Optional[CustomConfig] = Field(
        default=None, description="Stateful Ingestion Config"
    )


class StatefulLineageConfigMixin(LineageConfig):
    store_last_lineage_extraction_timestamp: bool = Field(
        default=False,
        description="Enable checking last lineage extraction date in store.",
    )

    @root_validator(pre=False)
    def lineage_stateful_option_validator(cls, values: Dict) -> Dict:
        sti = values.get("stateful_ingestion")
        if not sti or not sti.enabled:
            if values.get("store_last_lineage_extraction_timestamp"):
                logger.warning(
                    "Stateful ingestion is disabled, disabling store_last_lineage_extraction_timestamp config option as well"
                )
                values["store_last_lineage_extraction_timestamp"] = False

        return values


class StatefulProfilingConfigMixin(ConfigModel):
    store_last_profiling_timestamps: bool = Field(
        default=False,
        description="Enable storing last profile timestamp in store.",
    )

    @root_validator(pre=False)
    def profiling_stateful_option_validator(cls, values: Dict) -> Dict:
        sti = values.get("stateful_ingestion")
        if not sti or not sti.enabled:
            if values.get("store_last_profiling_timestamps"):
                logger.warning(
                    "Stateful ingestion is disabled, disabling store_last_profiling_timestamps config option as well"
                )
                values["store_last_profiling_timestamps"] = False
        return values


class StatefulUsageConfigMixin(BaseTimeWindowConfig):
    store_last_usage_extraction_timestamp: bool = Field(
        default=True,
        description="Enable checking last usage timestamp in store.",
    )

    @root_validator(pre=False)
    def last_usage_extraction_stateful_option_validator(cls, values: Dict) -> Dict:
        sti = values.get("stateful_ingestion")
        if not sti or not sti.enabled:
            if values.get("store_last_usage_extraction_timestamp"):
                logger.warning(
                    "Stateful ingestion is disabled, disabling store_last_usage_extraction_timestamp config option as well"
                )
                values["store_last_usage_extraction_timestamp"] = False
        return values


@dataclass
class StatefulIngestionReport(SourceReport):
    pass


class StatefulIngestionSourceBase(Source):
    """
    Defines the base class for all stateful sources.
    """

    def __init__(
        self, config: StatefulIngestionConfigBase, ctx: PipelineContext
    ) -> None:
        super().__init__(ctx)
        self.stateful_ingestion_config = config.stateful_ingestion
        self.last_checkpoints: Dict[JobId, Optional[Checkpoint]] = {}
        self.cur_checkpoints: Dict[JobId, Optional[Checkpoint]] = {}
        self.run_summaries_to_report: Dict[JobId, DatahubIngestionRunSummaryClass] = {}
        self.report: StatefulIngestionReport = StatefulIngestionReport()
        self._initialize_checkpointing_state_provider()
        self._usecase_handlers: Dict[JobId, StatefulIngestionUsecaseHandlerBase] = {}

    def warn(self, log: logging.Logger, key: str, reason: str) -> None:
        self.report.report_warning(key, reason)
        log.warning(f"{key} => {reason}")

    def error(self, log: logging.Logger, key: str, reason: str) -> None:
        self.report.report_failure(key, reason)
        log.error(f"{key} => {reason}")

    #
    # Checkpointing specific support.
    #

    def _initialize_checkpointing_state_provider(self) -> None:
        self.ingestion_checkpointing_state_provider: Optional[
            IngestionCheckpointingProviderBase
        ] = None
        if (
            self.stateful_ingestion_config is not None
            and self.stateful_ingestion_config.state_provider is not None
            and self.stateful_ingestion_config.enabled
        ):
            if self.ctx.pipeline_name is None:
                raise ConfigurationError(
                    "pipeline_name must be provided if stateful ingestion is enabled."
                )
            checkpointing_state_provider_class = (
                ingestion_checkpoint_provider_registry.get(
                    self.stateful_ingestion_config.state_provider.type
                )
            )
            if checkpointing_state_provider_class is None:
                raise ConfigurationError(
                    f"Cannot find checkpoint provider class of type={self.stateful_ingestion_config.state_provider.type} "
                    " in the registry! Please check the type of the checkpointing provider in your config."
                )
            config_dict: Dict[str, Any] = cast(
                Dict[str, Any],
                self.stateful_ingestion_config.state_provider.dict().get("config", {}),
            )
            self.ingestion_checkpointing_state_provider = (
                checkpointing_state_provider_class.create(
                    config_dict=config_dict,
                    ctx=self.ctx,
                    name=checkpointing_state_provider_class.__name__,
                )
            )
            assert self.ingestion_checkpointing_state_provider
            if self.stateful_ingestion_config.ignore_old_state:
                logger.warning(
                    "The 'ignore_old_state' config is True. The old checkpoint state will not be provided."
                )
            if self.stateful_ingestion_config.ignore_new_state:
                logger.warning(
                    "The 'ignore_new_state' config is True. The new checkpoint state will not be created."
                )
            # Add the checkpoint state provide to the platform context.
            self.ctx.register_checkpointer(self.ingestion_checkpointing_state_provider)

            logger.debug(
                f"Successfully created {self.stateful_ingestion_config.state_provider.type} state provider."
            )

    def register_stateful_ingestion_usecase_handler(
        self, usecase_handler: StatefulIngestionUsecaseHandlerBase
    ) -> None:
        """
        Registers a use-case handler with the common-base class.

        NOTE: The use-case handlers must invoke this method from their constructor(__init__) on the source.
        Also, the subclasses should not override this method.
        """
        assert usecase_handler.job_id not in self._usecase_handlers
        self._usecase_handlers[usecase_handler.job_id] = usecase_handler

    def is_stateful_ingestion_configured(self) -> bool:
        if (
            self.stateful_ingestion_config is not None
            and self.stateful_ingestion_config.enabled
            and self.ingestion_checkpointing_state_provider is not None
        ):
            return True
        return False

    def create_checkpoint(self, job_id: JobId) -> Optional[Checkpoint]:
        """
        Template base-class method that delegates the initial checkpoint state creation to
        the appropriate use-case handler registered for the job_id.
        """
        if job_id not in self._usecase_handlers:
            raise ValueError(f"No use-case handler for job_id{job_id}")
        return self._usecase_handlers[job_id].create_checkpoint()

    def is_checkpointing_enabled(self, job_id: JobId) -> bool:
        """
        Template base-class method that delegates the check if checkpointing is enabled for the job
        to the appropriate use-case handler registered for the job_id.
        """
        if job_id not in self._usecase_handlers:
            raise ValueError(f"No use-case handler for job_id{job_id}")
        return self._usecase_handlers[job_id].is_checkpointing_enabled()

    def get_platform_instance_id(self) -> Optional[str]:
        # This method is retained for backwards compatibility, but it is not
        # required that new sources implement it. We mainly need it for the
        # fallback logic in _get_last_checkpoint.
        raise NotImplementedError("no platform_instance_id configured")

    def _get_last_checkpoint(
        self, job_id: JobId, checkpoint_state_class: Type[StateType]
    ) -> Optional[Checkpoint]:
        """
        This is a template method implementation for querying the last checkpoint state.
        """
        last_checkpoint: Optional[Checkpoint] = None
        if self.is_stateful_ingestion_configured():
            # TRICKY: We currently don't include the platform_instance_id in the
            # checkpoint urn, but we previously did. As such, we need to fallback
            # and try the old urn format if the new format doesn't return anything.

            # Obtain the latest checkpoint from GMS for this job.
            assert self.ctx.pipeline_name
            last_checkpoint_aspect = self.ingestion_checkpointing_state_provider.get_latest_checkpoint(  # type: ignore
                pipeline_name=self.ctx.pipeline_name,
                job_name=job_id,
            )
            if last_checkpoint_aspect is None:
                # Try again with the platform_instance_id, if implemented.
                try:
                    platform_instance_id = self.get_platform_instance_id()
                except NotImplementedError:
                    pass
                else:
                    last_checkpoint_aspect = self.ingestion_checkpointing_state_provider.get_latest_checkpoint(  # type: ignore
                        pipeline_name=self.ctx.pipeline_name,
                        job_name=job_id,
                        platform_instance_id=platform_instance_id,
                    )

            # Convert it to a first-class Checkpoint object.
            last_checkpoint = Checkpoint[StateType].create_from_checkpoint_aspect(
                job_name=job_id,
                checkpoint_aspect=last_checkpoint_aspect,
                state_class=checkpoint_state_class,
            )
        return last_checkpoint

    # Base-class implementations for common state management tasks.
    def get_last_checkpoint(
        self, job_id: JobId, checkpoint_state_class: Type[StateType]
    ) -> Optional[Checkpoint]:
        if not self.is_stateful_ingestion_configured() or (
            self.stateful_ingestion_config
            and self.stateful_ingestion_config.ignore_old_state
        ):
            return None

        if job_id not in self.last_checkpoints:
            self.last_checkpoints[job_id] = self._get_last_checkpoint(
                job_id, checkpoint_state_class
            )
        return self.last_checkpoints[job_id]

    def get_current_checkpoint(self, job_id: JobId) -> Optional[Checkpoint]:
        if not self.is_stateful_ingestion_configured():
            return None

        if job_id not in self.cur_checkpoints:
            self.cur_checkpoints[job_id] = (
                self.create_checkpoint(job_id)
                if self.is_checkpointing_enabled(job_id)
                else None
            )
        return self.cur_checkpoints[job_id]

    def _prepare_checkpoint_states_for_commit(self) -> None:
        # Perform validations
        if not self.is_stateful_ingestion_configured():
            return None
        if (
            self.stateful_ingestion_config
            and self.stateful_ingestion_config.ignore_new_state
        ):
            logger.info(
                "The `ignore_new_state` config is True. Not committing current checkpoint."
            )
            return None
        if self.ctx.dry_run_mode or self.ctx.preview_mode:
            logger.warning(
                f"Will not be committing checkpoints in dry_run_mode(={self.ctx.dry_run_mode})"
                f" or preview_mode(={self.ctx.preview_mode})."
            )
            return None

        # Prepare the state the checkpointing provider should commit.
        job_checkpoint_aspects: Dict[JobId, DatahubIngestionCheckpointClass] = {}
        for job_name, job_checkpoint in self.cur_checkpoints.items():
            if job_checkpoint is None:
                continue
            job_checkpoint.prepare_for_commit()
            try:
                checkpoint_aspect = job_checkpoint.to_checkpoint_aspect(
                    self.stateful_ingestion_config.max_checkpoint_state_size  # type: ignore
                )
            except Exception as e:
                logger.error(
                    f"Failed to convert checkpoint to aspect for job {job_name}. It will not be committed.",
                    e,
                )
            else:
                if checkpoint_aspect is not None:
                    job_checkpoint_aspects[job_name] = checkpoint_aspect

        # Set the state to commit in the provider.
        assert self.ingestion_checkpointing_state_provider
        self.ingestion_checkpointing_state_provider.state_to_commit.update(
            job_checkpoint_aspects
        )

    def prepare_for_commit(self) -> None:
        """NOTE: Sources should call this method from their close method."""
        self._prepare_checkpoint_states_for_commit()

    def close(self) -> None:
        self.prepare_for_commit()
        super().close()



from collections import defaultdict
from typing import DefaultDict, Dict, Iterable

from datahub.ingestion.api.common import PipelineContext
from datahub.ingestion.api.decorators import (
    SourceCapability,
    SupportStatus,
    capability,
    config_class,
    platform_name,
    support_status,
)
from datahub.ingestion.api.source import Source
from datahub.ingestion.api.workunit import MetadataWorkUnit
from datahub.ingestion.source.aws.sagemaker_processors.common import (
    SagemakerSourceConfig,
    SagemakerSourceReport,
)
from datahub.ingestion.source.aws.sagemaker_processors.feature_groups import (
    FeatureGroupProcessor,
)
from datahub.ingestion.source.aws.sagemaker_processors.jobs import (
    JobKey,
    JobProcessor,
    ModelJob,
)
from datahub.ingestion.source.aws.sagemaker_processors.lineage import LineageProcessor
from datahub.ingestion.source.aws.sagemaker_processors.models import ModelProcessor


@platform_name("SageMaker")
@config_class(SagemakerSourceConfig)
@support_status(SupportStatus.CERTIFIED)
@capability(SourceCapability.LINEAGE_COARSE, "Enabled by default")
class SagemakerSource(Source):
    """
    This plugin extracts the following:

    - Feature groups
    - Models, jobs, and lineage between the two (e.g. when jobs output a model or a model is used by a job)
    """

    source_config: SagemakerSourceConfig
    report = SagemakerSourceReport()

    def __init__(self, config: SagemakerSourceConfig, ctx: PipelineContext):
        super().__init__(ctx)
        self.source_config = config
        self.report = SagemakerSourceReport()
        self.sagemaker_client = config.sagemaker_client
        self.env = config.env

    @classmethod
    def create(cls, config_dict, ctx):
        config = SagemakerSourceConfig.parse_obj(config_dict)
        return cls(config, ctx)

    def get_workunits(self) -> Iterable[MetadataWorkUnit]:
        # get common lineage graph
        lineage_processor = LineageProcessor(
            sagemaker_client=self.sagemaker_client, env=self.env, report=self.report
        )
        lineage = lineage_processor.get_lineage()

        # extract feature groups if specified
        if self.source_config.extract_feature_groups:
            feature_group_processor = FeatureGroupProcessor(
                sagemaker_client=self.sagemaker_client, env=self.env, report=self.report
            )
            yield from feature_group_processor.get_workunits()

        model_image_to_jobs: DefaultDict[str, Dict[JobKey, ModelJob]] = defaultdict(
            dict
        )
        model_name_to_jobs: DefaultDict[str, Dict[JobKey, ModelJob]] = defaultdict(dict)

        # extract jobs if specified
        if self.source_config.extract_jobs is not False:
            job_processor = JobProcessor(
                sagemaker_client=self.sagemaker_client,
                env=self.env,
                report=self.report,
                job_type_filter=self.source_config.extract_jobs,
                aws_region=self.source_config.aws_region,
            )
            yield from job_processor.get_workunits()

            model_image_to_jobs = job_processor.model_image_to_jobs
            model_name_to_jobs = job_processor.model_name_to_jobs

        # extract models if specified
        if self.source_config.extract_models:
            model_processor = ModelProcessor(
                sagemaker_client=self.sagemaker_client,
                env=self.env,
                report=self.report,
                model_image_to_jobs=model_image_to_jobs,
                model_name_to_jobs=model_name_to_jobs,
                lineage=lineage,
                aws_region=self.source_config.aws_region,
            )
            yield from model_processor.get_workunits()

    def get_report(self):
        return self.report



from textwrap import dedent
from typing import Optional

from pydantic.fields import Field
from pyhive.sqlalchemy_presto import PrestoDialect
from sqlalchemy import exc, sql
from sqlalchemy.engine import reflection

from datahub.ingestion.api.common import PipelineContext
from datahub.ingestion.api.decorators import (
    SourceCapability,
    SupportStatus,
    capability,
    config_class,
    platform_name,
    support_status,
)
from datahub.ingestion.source.sql.trino import (
    TrinoConfig,
    TrinoSource,
    _get_columns,
    get_table_comment,
    get_table_names,
)


# On Presto the information_schema.views does not return views but the tables table returns
@reflection.cache  # type: ignore
def get_view_names(self, connection, schema: str = None, **kw):  # type: ignore
    schema = schema or self._get_default_schema_name(connection)
    if schema is None:
        raise exc.NoSuchTableError("schema is required")
    query = dedent(
        """
        SELECT "table_name"
        FROM "information_schema"."tables"
        WHERE "table_schema" = :schema and "table_type" = 'VIEW'
    """
    ).strip()
    res = connection.execute(sql.text(query), schema=schema)
    return [row.table_name for row in res]


# The pyhive presto driver doesn't return view definitions, so we have to query it
@reflection.cache  # type: ignore
def get_view_definition(self, connection, view_name, schema=None, **kw):
    schema = schema or self._get_default_schema_name(connection)
    if schema is None:
        raise exc.NoSuchTableError("schema is required")

    if view_name is None:
        raise exc.NoSuchTableError("view_name is required")

    query = dedent(
        f"""
        SHOW CREATE VIEW "{schema}"."{view_name}"
    """
    ).strip()
    res = connection.execute(sql.text(query))
    return next(res)[0]


def _get_full_table(  # type: ignore
    self, table_name: str, schema: Optional[str] = None, quote: bool = True
) -> str:
    table_part = (
        self.identifier_preparer.quote_identifier(table_name) if quote else table_name
    )
    if schema:
        schema_part = (
            self.identifier_preparer.quote_identifier(schema) if quote else schema
        )
        return f"{schema_part}.{table_part}"

    return table_part


PrestoDialect.get_table_names = get_table_names
PrestoDialect.get_view_names = get_view_names
PrestoDialect.get_view_definition = get_view_definition
PrestoDialect.get_table_comment = get_table_comment
PrestoDialect.get_columns = _get_columns
PrestoDialect._get_full_table = _get_full_table


class PrestoConfig(TrinoConfig):
    # defaults
    scheme = Field(default="presto", description="", hidden_from_docs=True)


@platform_name("Presto", doc_order=1)
@config_class(PrestoConfig)
@support_status(SupportStatus.CERTIFIED)
@capability(SourceCapability.DOMAINS, "Supported via the `domain` config field")
@capability(SourceCapability.DATA_PROFILING, "Optionally enabled via configuration")
class PrestoSource(TrinoSource):
    """

    This plugin extracts the following:

    - Metadata for databases, schemas, and tables
    - Column types and schema associated with each table
    - Table, row, and column statistics via optional SQL profiling

    """

    config: PrestoConfig

    def __init__(self, config: PrestoConfig, ctx: PipelineContext):
        super().__init__(config, ctx, platform="presto")
        # We have to override platform as this source inherits from the Trino source

    @classmethod
    def create(cls, config_dict, ctx):
        config = PrestoConfig.parse_obj(config_dict)
        return cls(config, ctx)



from dataclasses import dataclass
from enum import Enum
from typing import Any, Dict, List, Optional

from datahub.emitter.mcp_builder import PlatformKey


class WorkspaceKey(PlatformKey):
    workspace: str


@dataclass
class Workspace:
    id: str
    name: str
    dashboards: List["Dashboard"]
    reports: List["Report"]
    datasets: Dict[str, "PowerBIDataset"]
    report_endorsements: Dict[str, List[str]]
    dashboard_endorsements: Dict[str, List[str]]
    scan_result: dict

    def get_urn_part(self):
        return self.name

    def get_workspace_key(self, platform_name: str) -> PlatformKey:
        return WorkspaceKey(
            workspace=self.get_urn_part(),
            platform=platform_name,
        )


@dataclass
class DataSource:
    id: str
    type: str
    raw_connection_detail: Dict

    def __members(self):
        return (self.id,)

    def __eq__(self, instance):
        return (
            isinstance(instance, DataSource)
            and self.__members() == instance.__members()
        )

    def __hash__(self):
        return hash(self.__members())


@dataclass
class Table:
    name: str
    full_name: str
    expression: Optional[str]

    # Pointer to the parent dataset.
    dataset: Optional["PowerBIDataset"] = None


@dataclass
class PowerBIDataset:
    id: str
    name: Optional[str]
    webUrl: Optional[str]
    workspace_id: str
    parameters: Optional[Dict[str, str]]

    # Table in datasets
    tables: List["Table"]
    tags: List[str]

    def get_urn_part(self):
        return f"datasets.{self.id}"

    def __members(self):
        return (self.id,)

    def __eq__(self, instance):
        return (
            isinstance(instance, PowerBIDataset)
            and self.__members() == instance.__members()
        )

    def __hash__(self):
        return hash(self.__members())


@dataclass
class Page:
    id: str
    displayName: str
    name: str
    order: int

    def get_urn_part(self):
        return f"pages.{self.id}"


@dataclass
class User:
    id: str
    displayName: str
    emailAddress: str
    graphId: str
    principalType: str

    def get_urn_part(self):
        return f"users.{self.id}"

    def __members(self):
        return (self.id,)

    def __eq__(self, instance):
        return isinstance(instance, User) and self.__members() == instance.__members()

    def __hash__(self):
        return hash(self.__members())


@dataclass
class Report:
    id: str
    name: str
    webUrl: Optional[str]
    embedUrl: str
    description: str
    dataset: Optional["PowerBIDataset"]
    pages: List["Page"]
    users: List["User"]
    tags: List[str]

    def get_urn_part(self):
        return f"reports.{self.id}"


@dataclass
class Tile:
    class CreatedFrom(Enum):
        REPORT = "Report"
        DATASET = "Dataset"
        VISUALIZATION = "Visualization"
        UNKNOWN = "UNKNOWN"

    id: str
    title: str
    embedUrl: str
    dataset: Optional["PowerBIDataset"]
    dataset_id: Optional[str]
    report: Optional[Report]
    createdFrom: CreatedFrom

    def get_urn_part(self):
        return f"charts.{self.id}"


@dataclass
class Dashboard:
    id: str
    displayName: str
    embedUrl: str
    webUrl: Optional[str]
    isReadOnly: Any
    workspace_id: str
    workspace_name: str
    tiles: List["Tile"]
    users: List["User"]
    tags: List[str]

    def get_urn_part(self):
        return f"dashboards.{self.id}"

    def __members(self):
        return (self.id,)

    def __eq__(self, instance):
        return (
            isinstance(instance, Dashboard) and self.__members() == instance.__members()
        )

    def __hash__(self):
        return hash(self.__members())


def new_powerbi_dataset(workspace_id: str, raw_instance: dict) -> PowerBIDataset:
    return PowerBIDataset(
        id=raw_instance["id"],
        name=raw_instance.get("name"),
        webUrl="{}/details".format(raw_instance.get("webUrl"))
        if raw_instance.get("webUrl") is not None
        else None,
        workspace_id=workspace_id,
        parameters=None,
        tables=[],
        tags=[],
    )



import logging
from typing import Any, Optional

from snowflake.connector import SnowflakeConnection
from snowflake.connector.cursor import DictCursor
from typing_extensions import Protocol

from datahub.configuration.common import MetaError
from datahub.configuration.pattern_utils import is_schema_allowed
from datahub.ingestion.source.snowflake.constants import (
    GENERIC_PERMISSION_ERROR_KEY,
    SNOWFLAKE_DEFAULT_CLOUD,
    SNOWFLAKE_REGION_CLOUD_REGION_MAPPING,
    SnowflakeObjectDomain,
)
from datahub.ingestion.source.snowflake.snowflake_config import SnowflakeV2Config
from datahub.ingestion.source.snowflake.snowflake_report import SnowflakeV2Report

logger: logging.Logger = logging.getLogger(__name__)


class SnowflakePermissionError(MetaError):
    """A permission error has happened"""


# Required only for mypy, since we are using mixin classes, and not inheritance.
# Reference - https://mypy.readthedocs.io/en/latest/more_types.html#mixin-classes
class SnowflakeLoggingProtocol(Protocol):
    logger: logging.Logger


class SnowflakeQueryProtocol(SnowflakeLoggingProtocol, Protocol):
    def get_connection(self) -> SnowflakeConnection:
        ...


class SnowflakeQueryMixin:
    def query(self: SnowflakeQueryProtocol, query: str) -> Any:
        try:
            self.logger.debug("Query : {}".format(query))
            resp = self.get_connection().cursor(DictCursor).execute(query)
            return resp

        except Exception as e:
            if is_permission_error(e):
                raise SnowflakePermissionError(e) from e
            raise


class SnowflakeCommonProtocol(SnowflakeLoggingProtocol, Protocol):
    config: SnowflakeV2Config
    report: SnowflakeV2Report

    def get_dataset_identifier(
        self, table_name: str, schema_name: str, db_name: str
    ) -> str:
        ...

    def get_dataset_identifier_from_qualified_name(self, qualified_name: str) -> str:
        ...

    def snowflake_identifier(self, identifier: str) -> str:
        ...

    def report_warning(self, key: str, reason: str) -> None:
        ...

    def report_error(self, key: str, reason: str) -> None:
        ...


class SnowflakeCommonMixin:
    platform = "snowflake"

    @staticmethod
    def create_snowsight_base_url(
        account_locator: str,
        cloud_region_id: str,
        cloud: str,
        privatelink: bool = False,
    ) -> Optional[str]:
        if privatelink:
            url = f"https://app.{account_locator}.{cloud_region_id}.privatelink.snowflakecomputing.com/"
        elif cloud == SNOWFLAKE_DEFAULT_CLOUD:
            url = f"https://app.snowflake.com/{cloud_region_id}/{account_locator}/"
        else:
            url = f"https://app.snowflake.com/{cloud_region_id}.{cloud}/{account_locator}/"
        return url

    @staticmethod
    def get_cloud_region_from_snowflake_region_id(region):
        if region in SNOWFLAKE_REGION_CLOUD_REGION_MAPPING.keys():
            cloud, cloud_region_id = SNOWFLAKE_REGION_CLOUD_REGION_MAPPING[region]
        elif region.startswith(("aws_", "gcp_", "azure_")):
            # e.g. aws_us_west_2, gcp_us_central1, azure_northeurope
            cloud, cloud_region_id = region.split("_", 1)
            cloud_region_id = cloud_region_id.replace("_", "-")
        else:
            raise Exception(f"Unknown snowflake region {region}")
        return cloud, cloud_region_id

    def _is_dataset_pattern_allowed(
        self: SnowflakeCommonProtocol,
        dataset_name: Optional[str],
        dataset_type: Optional[str],
    ) -> bool:
        if not dataset_type or not dataset_name:
            return True
        dataset_params = dataset_name.split(".")
        if dataset_type.lower() not in (
            SnowflakeObjectDomain.TABLE,
            SnowflakeObjectDomain.EXTERNAL_TABLE,
            SnowflakeObjectDomain.VIEW,
            SnowflakeObjectDomain.MATERIALIZED_VIEW,
        ):
            return False
        if len(dataset_params) != 3:
            self.report_warning(
                "invalid-dataset-pattern",
                f"Found {dataset_params} of type {dataset_type}",
            )
            # NOTE: this case returned `True` earlier when extracting lineage
            return False

        if not self.config.database_pattern.allowed(
            dataset_params[0].strip('"')
        ) or not is_schema_allowed(
            self.config.schema_pattern,
            dataset_params[1].strip('"'),
            dataset_params[0].strip('"'),
            self.config.match_fully_qualified_names,
        ):
            return False

        if dataset_type.lower() in {
            SnowflakeObjectDomain.TABLE
        } and not self.config.table_pattern.allowed(
            self.get_dataset_identifier_from_qualified_name(dataset_name)
        ):
            return False

        if dataset_type.lower() in {
            "view",
            "materialized_view",
        } and not self.config.view_pattern.allowed(
            self.get_dataset_identifier_from_qualified_name(dataset_name)
        ):
            return False

        return True

    def snowflake_identifier(self: SnowflakeCommonProtocol, identifier: str) -> str:
        # to be in in sync with older connector, convert name to lowercase
        if self.config.convert_urns_to_lowercase:
            return identifier.lower()
        return identifier

    @staticmethod
    def get_quoted_identifier_for_database(db_name):
        return f'"{db_name}"'

    @staticmethod
    def get_quoted_identifier_for_schema(db_name, schema_name):
        return f'"{db_name}"."{schema_name}"'

    @staticmethod
    def get_quoted_identifier_for_table(db_name, schema_name, table_name):
        return f'"{db_name}"."{schema_name}"."{table_name}"'

    def get_dataset_identifier(
        self: SnowflakeCommonProtocol, table_name: str, schema_name: str, db_name: str
    ) -> str:
        return self.snowflake_identifier(f"{db_name}.{schema_name}.{table_name}")

    # Qualified Object names from snowflake audit logs have quotes for for snowflake quoted identifiers,
    # For example "test-database"."test-schema".test_table
    # whereas we generate urns without quotes even for quoted identifiers for backward compatibility
    # and also unavailability of utility function to identify whether current table/schema/database
    # name should be quoted in above method get_dataset_identifier
    def get_dataset_identifier_from_qualified_name(
        self: SnowflakeCommonProtocol, qualified_name: str
    ) -> str:
        name_parts = qualified_name.split(".")
        if len(name_parts) != 3:
            self.report.report_warning(
                "invalid-dataset-pattern",
                f"Found non-parseable {name_parts} for {qualified_name}",
            )
            return self.snowflake_identifier(qualified_name.replace('"', ""))
        return self.get_dataset_identifier(
            name_parts[2].strip('"'), name_parts[1].strip('"'), name_parts[0].strip('"')
        )

    # Note - decide how to construct user urns.
    # Historically urns were created using part before @ from user's email.
    # Users without email were skipped from both user entries as well as aggregates.
    # However email is not mandatory field in snowflake user, user_name is always present.
    def get_user_identifier(
        self: SnowflakeCommonProtocol, user_name: str, user_email: Optional[str]
    ) -> str:
        if user_email:
            return user_email.split("@")[0]
        return self.snowflake_identifier(user_name)

    # TODO: Revisit this after stateful ingestion can commit checkpoint
    # for failures that do not affect the checkpoint
    def warn_if_stateful_else_error(
        self: SnowflakeCommonProtocol, key: str, reason: str
    ) -> None:
        if (
            self.config.stateful_ingestion is not None
            and self.config.stateful_ingestion.enabled
        ):
            self.report_warning(key, reason)
        else:
            self.report_error(key, reason)

    def report_warning(self: SnowflakeCommonProtocol, key: str, reason: str) -> None:
        self.report.report_warning(key, reason)
        self.logger.warning(f"{key} => {reason}")

    def report_error(self: SnowflakeCommonProtocol, key: str, reason: str) -> None:
        self.report.report_failure(key, reason)
        self.logger.error(f"{key} => {reason}")


class SnowflakeConnectionProtocol(SnowflakeLoggingProtocol, Protocol):
    connection: Optional[SnowflakeConnection]
    config: SnowflakeV2Config
    report: SnowflakeV2Report

    def create_connection(self) -> Optional[SnowflakeConnection]:
        ...

    def report_error(self, key: str, reason: str) -> None:
        ...


class SnowflakeConnectionMixin:
    def get_connection(self: SnowflakeConnectionProtocol) -> SnowflakeConnection:
        if self.connection is None:
            # Ideally this is never called here
            self.logger.info("Did you forget to initialize connection for module?")
            self.connection = self.create_connection()

        # Connection is already present by the time its used for query
        # Every module initializes the connection or fails and returns
        assert self.connection is not None
        return self.connection

    # If connection succeeds, return connection, else return None and report failure
    def create_connection(
        self: SnowflakeConnectionProtocol,
    ) -> Optional[SnowflakeConnection]:
        try:
            conn = self.config.get_connection()
        except Exception as e:
            logger.debug(e, exc_info=e)
            if "not granted to this user" in str(e):
                self.report_error(
                    GENERIC_PERMISSION_ERROR_KEY,
                    f"Failed to connect with snowflake due to error {e}",
                )
            else:
                logger.debug(e, exc_info=e)
                self.report_error(
                    "snowflake-connection",
                    f"Failed to connect to snowflake instance due to error {e}.",
                )
            return None
        else:
            return conn

    def close(self: SnowflakeConnectionProtocol) -> None:
        if self.connection is not None and not self.connection.is_closed():
            self.connection.close()


def is_permission_error(e: Exception) -> bool:
    msg = str(e)
    # 002003 (02000): SQL compilation error: Database/SCHEMA 'XXXX' does not exist or not authorized.
    # Insufficient privileges to operate on database 'XXXX'
    return "Insufficient privileges" in msg or "not authorized" in msg



import logging
from collections import defaultdict
from dataclasses import dataclass, field
from datetime import datetime
from functools import lru_cache
from typing import Dict, List, Optional

import pandas as pd
from snowflake.connector import SnowflakeConnection

from datahub.ingestion.source.snowflake.constants import SnowflakeObjectDomain
from datahub.ingestion.source.snowflake.snowflake_query import SnowflakeQuery
from datahub.ingestion.source.snowflake.snowflake_utils import SnowflakeQueryMixin
from datahub.ingestion.source.sql.sql_generic import BaseColumn, BaseTable, BaseView

logger: logging.Logger = logging.getLogger(__name__)


@dataclass
class SnowflakePK:
    name: str
    column_names: List[str]


@dataclass
class SnowflakeFK:
    name: str
    column_names: List[str]
    referred_database: str
    referred_schema: str
    referred_table: str
    referred_column_names: List[str]


@dataclass
class SnowflakeTag:
    database: str
    schema: str
    name: str
    value: str

    def identifier(self) -> str:
        return f"{self._id_prefix_as_str()}:{self.value}"

    def _id_prefix_as_str(self) -> str:
        return f"{self.database}.{self.schema}.{self.name}"


@dataclass
class SnowflakeColumn(BaseColumn):
    character_maximum_length: Optional[int]
    numeric_precision: Optional[int]
    numeric_scale: Optional[int]

    def get_precise_native_type(self):
        precise_native_type = self.data_type
        # https://docs.snowflake.com/en/sql-reference/data-types-numeric.html
        if (
            self.data_type in ("NUMBER", "NUMERIC", "DECIMAL")
            and self.numeric_precision is not None
            and self.numeric_scale is not None
        ):
            precise_native_type = (
                f"NUMBER({self.numeric_precision},{self.numeric_scale})"
            )
        # https://docs.snowflake.com/en/sql-reference/data-types-text.html
        elif (
            self.data_type in ("TEXT", "STRING", "VARCHAR")
            and self.character_maximum_length is not None
        ):
            precise_native_type = f"VARCHAR({self.character_maximum_length})"
        return precise_native_type


@dataclass
class SnowflakeTable(BaseTable):
    clustering_key: Optional[str] = None
    pk: Optional[SnowflakePK] = None
    columns: List[SnowflakeColumn] = field(default_factory=list)
    foreign_keys: List[SnowflakeFK] = field(default_factory=list)
    tags: Optional[List[SnowflakeTag]] = None
    column_tags: Dict[str, List[SnowflakeTag]] = field(default_factory=dict)
    sample_data: Optional[pd.DataFrame] = None


@dataclass
class SnowflakeView(BaseView):
    columns: List[SnowflakeColumn] = field(default_factory=list)
    tags: Optional[List[SnowflakeTag]] = None
    column_tags: Dict[str, List[SnowflakeTag]] = field(default_factory=dict)


@dataclass
class SnowflakeSchema:
    name: str
    created: Optional[datetime]
    last_altered: Optional[datetime]
    comment: Optional[str]
    tables: List[str] = field(default_factory=list)
    views: List[str] = field(default_factory=list)
    tags: Optional[List[SnowflakeTag]] = None


@dataclass
class SnowflakeDatabase:
    name: str
    created: Optional[datetime]
    comment: Optional[str]
    last_altered: Optional[datetime] = None
    schemas: List[SnowflakeSchema] = field(default_factory=list)
    tags: Optional[List[SnowflakeTag]] = None


class _SnowflakeTagCache:
    def __init__(self) -> None:
        # self._database_tags[<database_name>] = list of tags applied to database
        self._database_tags: Dict[str, List[SnowflakeTag]] = defaultdict(list)

        # self._schema_tags[<database_name>][<schema_name>] = list of tags applied to schema
        self._schema_tags: Dict[str, Dict[str, List[SnowflakeTag]]] = defaultdict(
            lambda: defaultdict(list)
        )

        # self._table_tags[<database_name>][<schema_name>][<table_name>] = list of tags applied to table
        self._table_tags: Dict[
            str, Dict[str, Dict[str, List[SnowflakeTag]]]
        ] = defaultdict(lambda: defaultdict(lambda: defaultdict(list)))

        # self._column_tags[<database_name>][<schema_name>][<table_name>][<column_name>] = list of tags applied to column
        self._column_tags: Dict[
            str, Dict[str, Dict[str, Dict[str, List[SnowflakeTag]]]]
        ] = defaultdict(
            lambda: defaultdict(lambda: defaultdict(lambda: defaultdict(list)))
        )

    def add_database_tag(self, db_name: str, tag: SnowflakeTag) -> None:
        self._database_tags[db_name].append(tag)

    def get_database_tags(self, db_name: str) -> List[SnowflakeTag]:
        return self._database_tags[db_name]

    def add_schema_tag(self, schema_name: str, db_name: str, tag: SnowflakeTag) -> None:
        self._schema_tags[db_name][schema_name].append(tag)

    def get_schema_tags(self, schema_name: str, db_name: str) -> List[SnowflakeTag]:
        return self._schema_tags.get(db_name, {}).get(schema_name, [])

    def add_table_tag(
        self, table_name: str, schema_name: str, db_name: str, tag: SnowflakeTag
    ) -> None:
        self._table_tags[db_name][schema_name][table_name].append(tag)

    def get_table_tags(
        self, table_name: str, schema_name: str, db_name: str
    ) -> List[SnowflakeTag]:
        return self._table_tags[db_name][schema_name][table_name]

    def add_column_tag(
        self,
        column_name: str,
        table_name: str,
        schema_name: str,
        db_name: str,
        tag: SnowflakeTag,
    ) -> None:
        self._column_tags[db_name][schema_name][table_name][column_name].append(tag)

    def get_column_tags_for_table(
        self, table_name: str, schema_name: str, db_name: str
    ) -> Dict[str, List[SnowflakeTag]]:
        return (
            self._column_tags.get(db_name, {}).get(schema_name, {}).get(table_name, {})
        )


class SnowflakeDataDictionary(SnowflakeQueryMixin):
    def __init__(self) -> None:
        self.logger = logger
        self.connection: Optional[SnowflakeConnection] = None

    def set_connection(self, connection: SnowflakeConnection) -> None:
        self.connection = connection

    def get_connection(self) -> SnowflakeConnection:
        # Connection is already present by the time this is called
        assert self.connection is not None
        return self.connection

    def show_databases(self) -> List[SnowflakeDatabase]:
        databases: List[SnowflakeDatabase] = []

        cur = self.query(
            SnowflakeQuery.show_databases(),
        )

        for database in cur:
            snowflake_db = SnowflakeDatabase(
                name=database["name"],
                created=database["created_on"],
                comment=database["comment"],
            )
            databases.append(snowflake_db)

        return databases

    def get_databases(self, db_name: str) -> List[SnowflakeDatabase]:
        databases: List[SnowflakeDatabase] = []

        cur = self.query(
            SnowflakeQuery.get_databases(db_name),
        )

        for database in cur:
            snowflake_db = SnowflakeDatabase(
                name=database["DATABASE_NAME"],
                created=database["CREATED"],
                last_altered=database["LAST_ALTERED"],
                comment=database["COMMENT"],
            )
            databases.append(snowflake_db)

        return databases

    def get_schemas_for_database(self, db_name: str) -> List[SnowflakeSchema]:
        snowflake_schemas = []

        cur = self.query(
            SnowflakeQuery.schemas_for_database(db_name),
        )

        for schema in cur:
            snowflake_schema = SnowflakeSchema(
                name=schema["SCHEMA_NAME"],
                created=schema["CREATED"],
                last_altered=schema["LAST_ALTERED"],
                comment=schema["COMMENT"],
            )
            snowflake_schemas.append(snowflake_schema)
        return snowflake_schemas

    @lru_cache(maxsize=1)
    def get_tables_for_database(
        self, db_name: str
    ) -> Optional[Dict[str, List[SnowflakeTable]]]:
        tables: Dict[str, List[SnowflakeTable]] = {}
        try:
            cur = self.query(
                SnowflakeQuery.tables_for_database(db_name),
            )
        except Exception as e:
            logger.debug(
                f"Failed to get all tables for database - {db_name}", exc_info=e
            )
            # Error - Information schema query returned too much data. Please repeat query with more selective predicates.
            return None

        for table in cur:
            if table["TABLE_SCHEMA"] not in tables:
                tables[table["TABLE_SCHEMA"]] = []
            tables[table["TABLE_SCHEMA"]].append(
                SnowflakeTable(
                    name=table["TABLE_NAME"],
                    created=table["CREATED"],
                    last_altered=table["LAST_ALTERED"],
                    size_in_bytes=table["BYTES"],
                    rows_count=table["ROW_COUNT"],
                    comment=table["COMMENT"],
                    clustering_key=table["CLUSTERING_KEY"],
                )
            )
        return tables

    def get_tables_for_schema(
        self, schema_name: str, db_name: str
    ) -> List[SnowflakeTable]:
        tables: List[SnowflakeTable] = []

        cur = self.query(
            SnowflakeQuery.tables_for_schema(schema_name, db_name),
        )

        for table in cur:
            tables.append(
                SnowflakeTable(
                    name=table["TABLE_NAME"],
                    created=table["CREATED"],
                    last_altered=table["LAST_ALTERED"],
                    size_in_bytes=table["BYTES"],
                    rows_count=table["ROW_COUNT"],
                    comment=table["COMMENT"],
                    clustering_key=table["CLUSTERING_KEY"],
                )
            )
        return tables

    @lru_cache(maxsize=1)
    def get_views_for_database(
        self, db_name: str
    ) -> Optional[Dict[str, List[SnowflakeView]]]:
        views: Dict[str, List[SnowflakeView]] = {}
        try:
            cur = self.query(SnowflakeQuery.show_views_for_database(db_name))
        except Exception as e:
            logger.debug(
                f"Failed to get all views for database - {db_name}", exc_info=e
            )
            # Error - Information schema query returned too much data. Please repeat query with more selective predicates.
            return None

        for table in cur:
            if table["schema_name"] not in views:
                views[table["schema_name"]] = []
            views[table["schema_name"]].append(
                SnowflakeView(
                    name=table["name"],
                    created=table["created_on"],
                    # last_altered=table["last_altered"],
                    comment=table["comment"],
                    view_definition=table["text"],
                    last_altered=table["created_on"],
                )
            )
        return views

    def get_views_for_schema(
        self, schema_name: str, db_name: str
    ) -> List[SnowflakeView]:
        views: List[SnowflakeView] = []

        cur = self.query(SnowflakeQuery.show_views_for_schema(schema_name, db_name))
        for table in cur:
            views.append(
                SnowflakeView(
                    name=table["name"],
                    created=table["created_on"],
                    # last_altered=table["last_altered"],
                    comment=table["comment"],
                    view_definition=table["text"],
                    last_altered=table["created_on"],
                )
            )
        return views

    @lru_cache(maxsize=1)
    def get_columns_for_schema(
        self, schema_name: str, db_name: str
    ) -> Optional[Dict[str, List[SnowflakeColumn]]]:
        columns: Dict[str, List[SnowflakeColumn]] = {}
        try:
            cur = self.query(SnowflakeQuery.columns_for_schema(schema_name, db_name))
        except Exception as e:
            logger.debug(
                f"Failed to get all columns for schema - {schema_name}", exc_info=e
            )
            # Error - Information schema query returned too much data.
            # Please repeat query with more selective predicates.
            return None

        for column in cur:
            if column["TABLE_NAME"] not in columns:
                columns[column["TABLE_NAME"]] = []
            columns[column["TABLE_NAME"]].append(
                SnowflakeColumn(
                    name=column["COLUMN_NAME"],
                    ordinal_position=column["ORDINAL_POSITION"],
                    is_nullable=column["IS_NULLABLE"] == "YES",
                    data_type=column["DATA_TYPE"],
                    comment=column["COMMENT"],
                    character_maximum_length=column["CHARACTER_MAXIMUM_LENGTH"],
                    numeric_precision=column["NUMERIC_PRECISION"],
                    numeric_scale=column["NUMERIC_SCALE"],
                )
            )
        return columns

    def get_columns_for_table(
        self, table_name: str, schema_name: str, db_name: str
    ) -> List[SnowflakeColumn]:
        columns: List[SnowflakeColumn] = []

        cur = self.query(
            SnowflakeQuery.columns_for_table(table_name, schema_name, db_name),
        )

        for column in cur:
            columns.append(
                SnowflakeColumn(
                    name=column["COLUMN_NAME"],
                    ordinal_position=column["ORDINAL_POSITION"],
                    is_nullable=column["IS_NULLABLE"] == "YES",
                    data_type=column["DATA_TYPE"],
                    comment=column["COMMENT"],
                    character_maximum_length=column["CHARACTER_MAXIMUM_LENGTH"],
                    numeric_precision=column["NUMERIC_PRECISION"],
                    numeric_scale=column["NUMERIC_SCALE"],
                )
            )
        return columns

    @lru_cache(maxsize=1)
    def get_pk_constraints_for_schema(
        self, schema_name: str, db_name: str
    ) -> Dict[str, SnowflakePK]:
        constraints: Dict[str, SnowflakePK] = {}
        cur = self.query(
            SnowflakeQuery.show_primary_keys_for_schema(schema_name, db_name),
        )

        for row in cur:
            if row["table_name"] not in constraints:
                constraints[row["table_name"]] = SnowflakePK(
                    name=row["constraint_name"], column_names=[]
                )
            constraints[row["table_name"]].column_names.append(row["column_name"])
        return constraints

    @lru_cache(maxsize=1)
    def get_fk_constraints_for_schema(
        self, schema_name: str, db_name: str
    ) -> Dict[str, List[SnowflakeFK]]:
        constraints: Dict[str, List[SnowflakeFK]] = {}
        fk_constraints_map: Dict[str, SnowflakeFK] = {}

        cur = self.query(
            SnowflakeQuery.show_foreign_keys_for_schema(schema_name, db_name),
        )

        for row in cur:
            if row["fk_name"] not in constraints:
                fk_constraints_map[row["fk_name"]] = SnowflakeFK(
                    name=row["fk_name"],
                    column_names=[],
                    referred_database=row["pk_database_name"],
                    referred_schema=row["pk_schema_name"],
                    referred_table=row["pk_table_name"],
                    referred_column_names=[],
                )

            if row["fk_table_name"] not in constraints:
                constraints[row["fk_table_name"]] = []

            fk_constraints_map[row["fk_name"]].column_names.append(
                row["fk_column_name"]
            )
            fk_constraints_map[row["fk_name"]].referred_column_names.append(
                row["pk_column_name"]
            )
            constraints[row["fk_table_name"]].append(fk_constraints_map[row["fk_name"]])

        return constraints

    def get_tags_for_database_without_propagation(
        self,
        db_name: str,
    ) -> _SnowflakeTagCache:
        cur = self.query(
            SnowflakeQuery.get_all_tags_in_database_without_propagation(db_name)
        )

        tags = _SnowflakeTagCache()

        for tag in cur:
            snowflake_tag = SnowflakeTag(
                database=tag["TAG_DATABASE"],
                schema=tag["TAG_SCHEMA"],
                name=tag["TAG_NAME"],
                value=tag["TAG_VALUE"],
            )

            # This is the name of the object, unless the object is a column, in which
            # case the name is in the `COLUMN_NAME` field.
            object_name = tag["OBJECT_NAME"]
            # This will be null if the object is a database or schema
            object_schema = tag["OBJECT_SCHEMA"]
            # This will be null if the object is a database
            object_database = tag["OBJECT_DATABASE"]

            domain = tag["DOMAIN"].lower()
            if domain == SnowflakeObjectDomain.DATABASE:
                tags.add_database_tag(object_name, snowflake_tag)
            elif domain == SnowflakeObjectDomain.SCHEMA:
                tags.add_schema_tag(object_name, object_database, snowflake_tag)
            elif domain == SnowflakeObjectDomain.TABLE:  # including views
                tags.add_table_tag(
                    object_name, object_schema, object_database, snowflake_tag
                )
            elif domain == SnowflakeObjectDomain.COLUMN:
                column_name = tag["COLUMN_NAME"]
                tags.add_column_tag(
                    column_name,
                    object_name,
                    object_schema,
                    object_database,
                    snowflake_tag,
                )
            else:
                # This should never happen.
                self.logger.error(f"Encountered an unexpected domain: {domain}")
                continue

        return tags

    def get_tags_for_object_with_propagation(
        self,
        domain: str,
        quoted_identifier: str,
        db_name: str,
    ) -> List[SnowflakeTag]:
        tags: List[SnowflakeTag] = []

        cur = self.query(
            SnowflakeQuery.get_all_tags_on_object_with_propagation(
                db_name, quoted_identifier, domain
            ),
        )

        for tag in cur:
            tags.append(
                SnowflakeTag(
                    database=tag["TAG_DATABASE"],
                    schema=tag["TAG_SCHEMA"],
                    name=tag["TAG_NAME"],
                    value=tag["TAG_VALUE"],
                )
            )
        return tags

    def get_tags_on_columns_for_table(
        self, quoted_table_name: str, db_name: str
    ) -> Dict[str, List[SnowflakeTag]]:
        tags: Dict[str, List[SnowflakeTag]] = defaultdict(list)
        cur = self.query(
            SnowflakeQuery.get_tags_on_columns_with_propagation(
                db_name, quoted_table_name
            ),
        )

        for tag in cur:
            column_name = tag["COLUMN_NAME"]
            snowflake_tag = SnowflakeTag(
                database=tag["TAG_DATABASE"],
                schema=tag["TAG_SCHEMA"],
                name=tag["TAG_NAME"],
                value=tag["TAG_VALUE"],
            )
            tags[column_name].append(snowflake_tag)

        return tags



from dataclasses import dataclass

from datahub.ingestion.api.report import EntityFilterReport
from datahub.ingestion.source.state.stale_entity_removal_handler import (
    StaleEntityRemovalSourceReport,
)


@dataclass
class UnityCatalogReport(StaleEntityRemovalSourceReport):
    metastores: EntityFilterReport = EntityFilterReport.field(type="metastore")
    catalogs: EntityFilterReport = EntityFilterReport.field(type="catalog")
    schemas: EntityFilterReport = EntityFilterReport.field(type="schema")
    tables: EntityFilterReport = EntityFilterReport.field(type="table/view")



import dataclasses
import logging
import os
import pathlib
import re
from collections import OrderedDict
from datetime import datetime
from typing import Any, Dict, Iterable, List, Optional, Tuple

import pydeequ
from pydeequ.analyzers import AnalyzerContext
from pyspark.conf import SparkConf
from pyspark.sql import SparkSession
from pyspark.sql.dataframe import DataFrame
from pyspark.sql.types import (
    ArrayType,
    BinaryType,
    BooleanType,
    ByteType,
    DateType,
    DecimalType,
    DoubleType,
    FloatType,
    IntegerType,
    LongType,
    MapType,
    NullType,
    ShortType,
    StringType,
    StructField,
    StructType,
    TimestampType,
)
from pyspark.sql.utils import AnalysisException
from smart_open import open as smart_open

import datahub.ingestion.source.s3.config
from datahub.emitter.mce_builder import (
    make_data_platform_urn,
    make_dataset_urn_with_platform_instance,
)
from datahub.emitter.mcp import MetadataChangeProposalWrapper
from datahub.ingestion.api.common import PipelineContext
from datahub.ingestion.api.decorators import (
    SourceCapability,
    SupportStatus,
    capability,
    config_class,
    platform_name,
    support_status,
)
from datahub.ingestion.api.source import Source, SourceReport
from datahub.ingestion.api.workunit import MetadataWorkUnit
from datahub.ingestion.source.aws.s3_boto_utils import get_s3_tags, list_folders
from datahub.ingestion.source.aws.s3_util import (
    get_bucket_name,
    get_bucket_relative_path,
    get_key_prefix,
    strip_s3_prefix,
)
from datahub.ingestion.source.s3.config import DataLakeSourceConfig, PathSpec
from datahub.ingestion.source.s3.data_lake_utils import ContainerWUCreator
from datahub.ingestion.source.s3.profiling import _SingleTableProfiler
from datahub.ingestion.source.s3.report import DataLakeSourceReport
from datahub.ingestion.source.schema_inference import avro, csv_tsv, json, parquet
from datahub.metadata.com.linkedin.pegasus2avro.common import Status
from datahub.metadata.com.linkedin.pegasus2avro.metadata.snapshot import DatasetSnapshot
from datahub.metadata.com.linkedin.pegasus2avro.mxe import MetadataChangeEvent
from datahub.metadata.com.linkedin.pegasus2avro.schema import (
    BooleanTypeClass,
    BytesTypeClass,
    DateTypeClass,
    NullTypeClass,
    NumberTypeClass,
    RecordTypeClass,
    SchemaFieldDataType,
    SchemaMetadata,
    StringTypeClass,
    TimeTypeClass,
)
from datahub.metadata.schema_classes import (
    ChangeTypeClass,
    DatasetPropertiesClass,
    MapTypeClass,
    OtherSchemaClass,
)
from datahub.telemetry import stats, telemetry
from datahub.utilities.perf_timer import PerfTimer

# hide annoying debug errors from py4j
logging.getLogger("py4j").setLevel(logging.ERROR)
logger: logging.Logger = logging.getLogger(__name__)

# for a list of all types, see https://spark.apache.org/docs/3.0.3/api/python/_modules/pyspark/sql/types.html
_field_type_mapping = {
    NullType: NullTypeClass,
    StringType: StringTypeClass,
    BinaryType: BytesTypeClass,
    BooleanType: BooleanTypeClass,
    DateType: DateTypeClass,
    TimestampType: TimeTypeClass,
    DecimalType: NumberTypeClass,
    DoubleType: NumberTypeClass,
    FloatType: NumberTypeClass,
    ByteType: BytesTypeClass,
    IntegerType: NumberTypeClass,
    LongType: NumberTypeClass,
    ShortType: NumberTypeClass,
    ArrayType: NullTypeClass,
    MapType: MapTypeClass,
    StructField: RecordTypeClass,
    StructType: RecordTypeClass,
}
SAMPLE_SIZE = 100
PAGE_SIZE = 1000


def get_column_type(
    report: SourceReport, dataset_name: str, column_type: str
) -> SchemaFieldDataType:
    """
    Maps known Spark types to datahub types
    """
    TypeClass: Any = None

    for field_type, type_class in _field_type_mapping.items():
        if isinstance(column_type, field_type):
            TypeClass = type_class
            break

    # if still not found, report the warning
    if TypeClass is None:
        report.report_warning(
            dataset_name, f"unable to map type {column_type} to metadata schema"
        )
        TypeClass = NullTypeClass

    return SchemaFieldDataType(type=TypeClass())


# config flags to emit telemetry for
config_options_to_report = [
    "platform",
    "use_relative_path",
    "ignore_dotfiles",
]

# profiling flags to emit telemetry for
profiling_flags_to_report = [
    "profile_table_level_only",
    "include_field_null_count",
    "include_field_min_value",
    "include_field_max_value",
    "include_field_mean_value",
    "include_field_median_value",
    "include_field_stddev_value",
    "include_field_quantiles",
    "include_field_distinct_value_frequencies",
    "include_field_histogram",
    "include_field_sample_values",
]


# LOCAL_BROWSE_PATH_TRANSFORMER_CONFIG = AddDatasetBrowsePathConfig(
#     path_templates=["/ENV/PLATFORMDATASET_PARTS"], replace_existing=True
# )
#
# LOCAL_BROWSE_PATH_TRANSFORMER = AddDatasetBrowsePathTransformer(
#     ctx=None, config=LOCAL_BROWSE_PATH_TRANSFORMER_CONFIG
# )


@dataclasses.dataclass
class TableData:
    display_name: str
    is_s3: bool
    full_path: str
    partitions: Optional[OrderedDict]
    timestamp: datetime
    table_path: str
    size_in_bytes: int
    number_of_files: int


@platform_name("S3 Data Lake", id="s3")
@config_class(DataLakeSourceConfig)
@support_status(SupportStatus.INCUBATING)
@capability(SourceCapability.DATA_PROFILING, "Optionally enabled via configuration")
@capability(SourceCapability.TAGS, "Can extract S3 object/bucket tags if enabled")
class S3Source(Source):
    """
    This plugin extracts:

    - Row and column counts for each table
    - For each column, if profiling is enabled:
      - null counts and proportions
      - distinct counts and proportions
      - minimum, maximum, mean, median, standard deviation, some quantile values
      - histograms or frequencies of unique values

    This connector supports both local files as well as those stored on AWS S3 (which must be identified using the prefix `s3://`). Supported file types are as follows:

    - CSV
    - TSV
    - JSON
    - Parquet
    - Apache Avro

    Schemas for Parquet and Avro files are extracted as provided.

    Schemas for schemaless formats (CSV, TSV, JSON) are inferred. For CSV and TSV files, we consider the first 100 rows by default, which can be controlled via the `max_rows` recipe parameter (see [below](#config-details))
    JSON file schemas are inferred on the basis of the entire file (given the difficulty in extracting only the first few objects of the file), which may impact performance.
    We are working on using iterator-based JSON parsers to avoid reading in the entire JSON object.

    Note that because the profiling is run with PySpark, we require Spark 3.0.3 with Hadoop 3.2 to be installed (see [compatibility](#compatibility) for more details). If profiling, make sure that permissions for **s3a://** access are set because Spark and Hadoop use the s3a:// protocol to interface with AWS (schema inference outside of profiling requires s3:// access).
    Enabling profiling will slow down ingestion runs.
    """

    source_config: DataLakeSourceConfig
    report: DataLakeSourceReport
    profiling_times_taken: List[float]
    container_WU_creator: ContainerWUCreator

    def __init__(self, config: DataLakeSourceConfig, ctx: PipelineContext):
        super().__init__(ctx)
        self.source_config = config
        self.report = DataLakeSourceReport()
        self.profiling_times_taken = []
        config_report = {
            config_option: config.dict().get(config_option)
            for config_option in config_options_to_report
        }
        config_report = {**config_report, "profiling_enabled": config.profiling.enabled}

        telemetry.telemetry_instance.ping(
            "data_lake_config",
            config_report,
        )

        if config.profiling.enabled:
            telemetry.telemetry_instance.ping(
                "data_lake_profiling_config",
                {
                    config_flag: config.profiling.dict().get(config_flag)
                    for config_flag in profiling_flags_to_report
                },
            )
            self.init_spark()

    def init_spark(self):
        conf = SparkConf()

        conf.set(
            "spark.jars.packages",
            ",".join(
                [
                    "org.apache.hadoop:hadoop-aws:3.0.3",
                    "org.apache.spark:spark-avro_2.12:3.0.3",
                    pydeequ.deequ_maven_coord,
                ]
            ),
        )

        if self.source_config.aws_config is not None:
            credentials = self.source_config.aws_config.get_credentials()

            aws_access_key_id = credentials.get("aws_access_key_id")
            aws_secret_access_key = credentials.get("aws_secret_access_key")
            aws_session_token = credentials.get("aws_session_token")

            aws_provided_credentials = [
                aws_access_key_id,
                aws_secret_access_key,
                aws_session_token,
            ]

            if any(x is not None for x in aws_provided_credentials):
                # see https://hadoop.apache.org/docs/r3.0.3/hadoop-aws/tools/hadoop-aws/index.html#Changing_Authentication_Providers
                if all(x is not None for x in aws_provided_credentials):
                    conf.set(
                        "spark.hadoop.fs.s3a.aws.credentials.provider",
                        "org.apache.hadoop.fs.s3a.TemporaryAWSCredentialsProvider",
                    )

                else:
                    conf.set(
                        "spark.hadoop.fs.s3a.aws.credentials.provider",
                        "org.apache.hadoop.fs.s3a.SimpleAWSCredentialsProvider",
                    )

                if aws_access_key_id is not None:
                    conf.set("spark.hadoop.fs.s3a.access.key", aws_access_key_id)
                if aws_secret_access_key is not None:
                    conf.set(
                        "spark.hadoop.fs.s3a.secret.key",
                        aws_secret_access_key,
                    )
                if aws_session_token is not None:
                    conf.set(
                        "spark.hadoop.fs.s3a.session.token",
                        aws_session_token,
                    )
            else:
                # if no explicit AWS config is provided, use a default AWS credentials provider
                conf.set(
                    "spark.hadoop.fs.s3a.aws.credentials.provider",
                    "org.apache.hadoop.fs.s3a.AnonymousAWSCredentialsProvider",
                )

            if self.source_config.aws_config.aws_endpoint_url is not None:
                conf.set(
                    "fs.s3a.endpoint", self.source_config.aws_config.aws_endpoint_url
                )
            if self.source_config.aws_config.aws_region is not None:
                conf.set(
                    "fs.s3a.endpoint.region", self.source_config.aws_config.aws_region
                )

        conf.set("spark.jars.excludes", pydeequ.f2j_maven_coord)
        conf.set("spark.driver.memory", self.source_config.spark_driver_memory)

        self.spark = SparkSession.builder.config(conf=conf).getOrCreate()

    @classmethod
    def create(cls, config_dict, ctx):
        config = DataLakeSourceConfig.parse_obj(config_dict)

        return cls(config, ctx)

    def read_file_spark(self, file: str, ext: str) -> Optional[DataFrame]:
        logger.debug(f"Opening file {file} for profiling in spark")
        file = file.replace("s3://", "s3a://")

        telemetry.telemetry_instance.ping("data_lake_file", {"extension": ext})

        if ext.endswith(".parquet"):
            df = self.spark.read.parquet(file)
        elif ext.endswith(".csv"):
            # see https://sparkbyexamples.com/pyspark/pyspark-read-csv-file-into-dataframe
            df = self.spark.read.csv(
                file,
                header="True",
                inferSchema="True",
                sep=",",
                ignoreLeadingWhiteSpace=True,
                ignoreTrailingWhiteSpace=True,
            )
        elif ext.endswith(".tsv"):
            df = self.spark.read.csv(
                file,
                header="True",
                inferSchema="True",
                sep="\t",
                ignoreLeadingWhiteSpace=True,
                ignoreTrailingWhiteSpace=True,
            )
        elif ext.endswith(".json"):
            df = self.spark.read.json(file)
        elif ext.endswith(".avro"):
            try:
                df = self.spark.read.format("avro").load(file)
            except AnalysisException:
                self.report.report_warning(
                    file,
                    "To ingest avro files, please install the spark-avro package: https://mvnrepository.com/artifact/org.apache.spark/spark-avro_2.12/3.0.3",
                )
                return None

        # TODO: add support for more file types
        # elif file.endswith(".orc"):
        # df = self.spark.read.orc(file)
        else:
            self.report.report_warning(file, f"file {file} has unsupported extension")
            return None
        logger.debug(f"dataframe read for file {file} with row count {df.count()}")
        # replace periods in names because they break PyDeequ
        # see https://mungingdata.com/pyspark/avoid-dots-periods-column-names/
        return df.toDF(*(c.replace(".", "_") for c in df.columns))

    def get_fields(self, table_data: TableData, path_spec: PathSpec) -> List:
        if table_data.is_s3:
            if self.source_config.aws_config is None:
                raise ValueError("AWS config is required for S3 file sources")

            s3_client = self.source_config.aws_config.get_s3_client(
                self.source_config.verify_ssl
            )

            file = smart_open(
                table_data.full_path, "rb", transport_params={"client": s3_client}
            )
        else:
            file = open(table_data.full_path, "rb")

        fields = []

        extension = pathlib.Path(table_data.full_path).suffix
        if path_spec.enable_compression and (
            extension[1:]
            in datahub.ingestion.source.aws.path_spec.SUPPORTED_COMPRESSIONS
        ):
            # Removing the compression extension and using the one before that like .json.gz -> .json
            extension = pathlib.Path(table_data.full_path).with_suffix("").suffix
        if extension == "" and path_spec.default_extension:
            extension = f".{path_spec.default_extension}"

        try:
            if extension == ".parquet":
                fields = parquet.ParquetInferrer().infer_schema(file)
            elif extension == ".csv":
                fields = csv_tsv.CsvInferrer(
                    max_rows=self.source_config.max_rows
                ).infer_schema(file)
            elif extension == ".tsv":
                fields = csv_tsv.TsvInferrer(
                    max_rows=self.source_config.max_rows
                ).infer_schema(file)
            elif extension == ".json":
                fields = json.JsonInferrer().infer_schema(file)
            elif extension == ".avro":
                fields = avro.AvroInferrer().infer_schema(file)
            else:
                self.report.report_warning(
                    table_data.full_path,
                    f"file {table_data.full_path} has unsupported extension",
                )
            file.close()
        except Exception as e:
            self.report.report_warning(
                table_data.full_path,
                f"could not infer schema for file {table_data.full_path}: {e}",
            )
            file.close()
        logger.debug(f"Extracted fields in schema: {fields}")
        fields = sorted(fields, key=lambda f: f.fieldPath)

        return fields

    def get_table_profile(
        self, table_data: TableData, dataset_urn: str
    ) -> Iterable[MetadataWorkUnit]:
        # read in the whole table with Spark for profiling
        table = None
        try:
            if table_data.partitions:
                table = self.read_file_spark(
                    table_data.table_path, os.path.splitext(table_data.full_path)[1]
                )
            else:
                table = self.read_file_spark(
                    table_data.full_path, os.path.splitext(table_data.full_path)[1]
                )
        except Exception as e:
            logger.error(e)

        # if table is not readable, skip
        if table is None:
            self.report.report_warning(
                table_data.display_name,
                f"unable to read table {table_data.display_name} from file {table_data.full_path}",
            )
            return

        with PerfTimer() as timer:
            # init PySpark analysis object
            logger.debug(
                f"Profiling {table_data.full_path}: reading file and computing nulls+uniqueness {datetime.now().strftime('%d/%m/%Y %H:%M:%S')}"
            )
            table_profiler = _SingleTableProfiler(
                table,
                self.spark,
                self.source_config.profiling,
                self.report,
                table_data.full_path,
            )

            logger.debug(
                f"Profiling {table_data.full_path}: preparing profilers to run {datetime.now().strftime('%d/%m/%Y %H:%M:%S')}"
            )
            # instead of computing each profile individually, we run them all in a single analyzer.run() call
            # we use a single call because the analyzer optimizes the number of calls to the underlying profiler
            # since multiple profiles reuse computations, this saves a lot of time
            table_profiler.prepare_table_profiles()

            # compute the profiles
            logger.debug(
                f"Profiling {table_data.full_path}: computing profiles {datetime.now().strftime('%d/%m/%Y %H:%M:%S')}"
            )
            analysis_result = table_profiler.analyzer.run()
            analysis_metrics = AnalyzerContext.successMetricsAsDataFrame(
                self.spark, analysis_result
            )

            logger.debug(
                f"Profiling {table_data.full_path}: extracting profiles {datetime.now().strftime('%d/%m/%Y %H:%M:%S')}"
            )
            table_profiler.extract_table_profiles(analysis_metrics)

            time_taken = timer.elapsed_seconds()

            logger.info(
                f"Finished profiling {table_data.full_path}; took {time_taken:.3f} seconds"
            )

            self.profiling_times_taken.append(time_taken)

        mcp = MetadataChangeProposalWrapper(
            entityType="dataset",
            entityUrn=dataset_urn,
            changeType=ChangeTypeClass.UPSERT,
            aspectName="datasetProfile",
            aspect=table_profiler.profile,
        )
        wu = MetadataWorkUnit(
            id=f"profile-{self.source_config.platform}-{table_data.table_path}", mcp=mcp
        )
        self.report.report_workunit(wu)
        yield wu

    def ingest_table(
        self, table_data: TableData, path_spec: PathSpec
    ) -> Iterable[MetadataWorkUnit]:
        logger.info(f"Extracting table schema from file: {table_data.full_path}")
        browse_path: str = (
            strip_s3_prefix(table_data.table_path)
            if table_data.is_s3
            else table_data.table_path.strip("/")
        )

        data_platform_urn = make_data_platform_urn(self.source_config.platform)
        logger.info(f"Creating dataset urn with name: {browse_path}")
        dataset_urn = make_dataset_urn_with_platform_instance(
            self.source_config.platform,
            browse_path,
            self.source_config.platform_instance,
            self.source_config.env,
        )

        dataset_snapshot = DatasetSnapshot(
            urn=dataset_urn,
            aspects=[Status(removed=False)],
        )

        customProperties: Optional[Dict[str, str]] = None
        if not path_spec.sample_files:
            customProperties = {
                "number_of_files": str(table_data.number_of_files),
                "size_in_bytes": str(table_data.size_in_bytes),
            }
            if table_data.is_s3:
                customProperties["table_path"] = str(table_data.table_path)

        dataset_properties = DatasetPropertiesClass(
            description="",
            name=table_data.display_name,
            customProperties=customProperties,
        )
        dataset_snapshot.aspects.append(dataset_properties)

        fields = self.get_fields(table_data, path_spec)
        schema_metadata = SchemaMetadata(
            schemaName=table_data.display_name,
            platform=data_platform_urn,
            version=0,
            hash="",
            fields=fields,
            platformSchema=OtherSchemaClass(rawSchema=""),
        )
        dataset_snapshot.aspects.append(schema_metadata)
        if (
            self.source_config.use_s3_bucket_tags
            or self.source_config.use_s3_object_tags
        ):
            bucket = get_bucket_name(table_data.table_path)
            key_prefix = (
                get_key_prefix(table_data.table_path)
                if table_data.full_path == table_data.table_path
                else None
            )
            s3_tags = get_s3_tags(
                bucket,
                key_prefix,
                dataset_urn,
                self.source_config.aws_config,
                self.ctx,
                self.source_config.use_s3_bucket_tags,
                self.source_config.use_s3_object_tags,
                self.source_config.verify_ssl,
            )
            if s3_tags is not None:
                dataset_snapshot.aspects.append(s3_tags)

        mce = MetadataChangeEvent(proposedSnapshot=dataset_snapshot)
        wu = MetadataWorkUnit(id=table_data.table_path, mce=mce)
        self.report.report_workunit(wu)
        yield wu

        container_wus = self.container_WU_creator.create_container_hierarchy(
            table_data.table_path, table_data.is_s3, dataset_urn
        )
        for wu in container_wus:
            self.report.report_workunit(wu)
            yield wu

        if self.source_config.profiling.enabled:
            yield from self.get_table_profile(table_data, dataset_urn)

    def get_prefix(self, relative_path: str) -> str:
        index = re.search(r"[\*|\{]", relative_path)
        if index:
            return relative_path[: index.start()]
        else:
            return relative_path

    def extract_table_name(self, path_spec: PathSpec, named_vars: dict) -> str:
        if path_spec.table_name is None:
            raise ValueError("path_spec.table_name is not set")
        return path_spec.table_name.format_map(named_vars)

    def extract_table_data(
        self, path_spec: PathSpec, path: str, timestamp: datetime, size: int
    ) -> TableData:
        logger.debug(f"Getting table data for path: {path}")
        table_name, table_path = path_spec.extract_table_name_and_path(path)
        table_data = None
        table_data = TableData(
            display_name=table_name,
            is_s3=path_spec.is_s3,
            full_path=path,
            partitions=None,
            timestamp=timestamp,
            table_path=table_path,
            number_of_files=1,
            size_in_bytes=size,
        )
        return table_data

    def resolve_templated_folders(self, bucket_name: str, prefix: str) -> Iterable[str]:
        folder_split: List[str] = prefix.split("*", 1)
        # If the len of split is 1 it means we don't have * in the prefix
        if len(folder_split) == 1:
            yield prefix
            return

        folders: Iterable[str] = list_folders(
            bucket_name, folder_split[0], self.source_config.aws_config
        )
        for folder in folders:
            yield from self.resolve_templated_folders(
                bucket_name, f"{folder}{folder_split[1]}"
            )

    def s3_browser(self, path_spec: PathSpec) -> Iterable[Tuple[str, datetime, int]]:
        if self.source_config.aws_config is None:
            raise ValueError("aws_config not set. Cannot browse s3")
        s3 = self.source_config.aws_config.get_s3_resource(
            self.source_config.verify_ssl
        )
        bucket_name = get_bucket_name(path_spec.include)
        logger.debug(f"Scanning bucket: {bucket_name}")
        bucket = s3.Bucket(bucket_name)
        prefix = self.get_prefix(get_bucket_relative_path(path_spec.include))
        logger.debug(f"Scanning objects with prefix:{prefix}")
        matches = re.finditer(r"{\s*\w+\s*}", path_spec.include, re.MULTILINE)
        matches_list = list(matches)
        if matches_list and path_spec.sample_files:
            # Replace the patch_spec include's templates with star because later we want to resolve all the stars
            # to actual directories.
            # For example:
            # "s3://my-test-bucket/*/{dept}/*/{table}/*/*.*" -> "s3://my-test-bucket/*/*/*/{table}/*/*.*"
            # We only keep the last template as a marker to know the point util we need to resolve path.
            # After the marker we can safely get sample files for sampling because it is not used in the
            # table name, so we don't need all the files.
            # This speed up processing but we won't be able to get a precise modification date/size/number of files.
            max_start: int = -1
            include: str = path_spec.include
            max_match: str = ""
            for match in matches_list:
                pos = include.find(match.group())
                if pos > max_start:
                    if max_match:
                        include = include.replace(max_match, "*")
                    max_start = match.start()
                    max_match = match.group()

            table_index = include.find(max_match)
            for folder in self.resolve_templated_folders(
                bucket_name, get_bucket_relative_path(include[:table_index])
            ):
                for f in list_folders(
                    bucket_name, f"{folder}", self.source_config.aws_config
                ):
                    logger.info(f"Processing folder: {f}")

                    for obj in (
                        bucket.objects.filter(Prefix=f"{f}")
                        .page_size(PAGE_SIZE)
                        .limit(SAMPLE_SIZE)
                    ):
                        s3_path = f"s3://{obj.bucket_name}/{obj.key}"
                        logger.debug(f"Samping file: {s3_path}")
                        yield s3_path, obj.last_modified, obj.size,
        else:
            logger.debug(
                "No template in the pathspec can't do sampling, fallbacking to do full scan"
            )
            path_spec.sample_files = False
            for obj in bucket.objects.filter(Prefix=prefix).page_size(PAGE_SIZE):
                s3_path = f"s3://{obj.bucket_name}/{obj.key}"
                logger.debug(f"Path: {s3_path}")
                yield s3_path, obj.last_modified, obj.size,

    def local_browser(self, path_spec: PathSpec) -> Iterable[Tuple[str, datetime, int]]:
        prefix = self.get_prefix(path_spec.include)
        if os.path.isfile(prefix):
            logger.debug(f"Scanning single local file: {prefix}")
            yield prefix, datetime.utcfromtimestamp(
                os.path.getmtime(prefix)
            ), os.path.getsize(prefix)
        else:
            logger.debug(f"Scanning files under local folder: {prefix}")
            for root, dirs, files in os.walk(prefix):
                for file in sorted(files):
                    full_path = os.path.join(root, file)
                    yield full_path, datetime.utcfromtimestamp(
                        os.path.getmtime(full_path)
                    ), os.path.getsize(full_path)

    def get_workunits(self) -> Iterable[MetadataWorkUnit]:
        self.container_WU_creator = ContainerWUCreator(
            self.source_config.platform,
            self.source_config.platform_instance,
            self.source_config.env,
        )
        with PerfTimer() as timer:
            assert self.source_config.path_specs
            for path_spec in self.source_config.path_specs:
                file_browser = (
                    self.s3_browser(path_spec)
                    if self.source_config.platform == "s3"
                    else self.local_browser(path_spec)
                )
                table_dict: Dict[str, TableData] = {}
                for file, timestamp, size in file_browser:
                    if not path_spec.allowed(file):
                        continue
                    table_data = self.extract_table_data(
                        path_spec, file, timestamp, size
                    )
                    if table_data.table_path not in table_dict:
                        table_dict[table_data.table_path] = table_data
                    else:
                        logger.debug(
                            f"Update schema on partition file updates is set to: {self.source_config.update_schema_on_partition_file_updates!s}"
                        )
                        if (
                            self.source_config.update_schema_on_partition_file_updates
                            and not path_spec.sample_files
                        ):
                            logger.info(
                                "Will update table schema as file within the partitions has an updated schema."
                            )
                            table_dict[table_data.table_path] = table_data
                        table_dict[table_data.table_path].number_of_files = (
                            table_dict[table_data.table_path].number_of_files + 1
                        )
                        table_dict[table_data.table_path].size_in_bytes = (
                            table_dict[table_data.table_path].size_in_bytes
                            + table_data.size_in_bytes
                        )
                        if (
                            table_dict[table_data.table_path].timestamp
                            < table_data.timestamp
                        ):
                            table_dict[
                                table_data.table_path
                            ].full_path = table_data.full_path
                            table_dict[
                                table_data.table_path
                            ].timestamp = table_data.timestamp

                for guid, table_data in table_dict.items():
                    yield from self.ingest_table(table_data, path_spec)

            if not self.source_config.profiling.enabled:
                return

            total_time_taken = timer.elapsed_seconds()

            logger.info(
                f"Profiling {len(self.profiling_times_taken)} table(s) finished in {total_time_taken:.3f} seconds"
            )

            time_percentiles: Dict[str, float] = {}

            if len(self.profiling_times_taken) > 0:
                percentiles = [50, 75, 95, 99]
                percentile_values = stats.calculate_percentiles(
                    self.profiling_times_taken, percentiles
                )

                time_percentiles = {
                    f"table_time_taken_p{percentile}": stats.discretize(
                        percentile_values[percentile]
                    )
                    for percentile in percentiles
                }

            telemetry.telemetry_instance.ping(
                "data_lake_profiling_summary",
                # bucket by taking floor of log of time taken
                {
                    "total_time_taken": stats.discretize(total_time_taken),
                    "count": stats.discretize(len(self.profiling_times_taken)),
                    "platform": self.source_config.platform,
                    **time_percentiles,
                },
            )

    def get_report(self):
        return self.report



import datetime
from abc import ABCMeta, abstractmethod
from collections import defaultdict
from dataclasses import dataclass, field
from enum import Enum
from typing import Dict, Generic, Iterable, Optional, Set, Type, TypeVar, Union, cast

from pydantic import BaseModel

from datahub.configuration.common import ConfigModel
from datahub.emitter.mcp_builder import mcps_from_mce
from datahub.ingestion.api.closeable import Closeable
from datahub.ingestion.api.common import PipelineContext, RecordEnvelope, WorkUnit
from datahub.ingestion.api.report import Report
from datahub.ingestion.api.workunit import MetadataWorkUnit
from datahub.metadata.com.linkedin.pegasus2avro.mxe import MetadataChangeEvent
from datahub.utilities.lossy_collections import LossyDict, LossyList
from datahub.utilities.type_annotations import get_class_from_annotation


class SourceCapability(Enum):
    PLATFORM_INSTANCE = "Platform Instance"
    DOMAINS = "Domains"
    DATA_PROFILING = "Data Profiling"
    USAGE_STATS = "Dataset Usage"
    PARTITION_SUPPORT = "Partition Support"
    DESCRIPTIONS = "Descriptions"
    LINEAGE_COARSE = "Table-Level Lineage"
    LINEAGE_FINE = "Column-level Lineage"
    OWNERSHIP = "Extract Ownership"
    DELETION_DETECTION = "Detect Deleted Entities"
    TAGS = "Extract Tags"
    SCHEMA_METADATA = "Schema Metadata"
    CONTAINERS = "Asset Containers"


@dataclass
class SourceReport(Report):
    events_produced: int = 0
    events_produced_per_sec: int = 0

    _urns_seen: Set[str] = field(default_factory=set)
    entities: Dict[str, list] = field(default_factory=lambda: defaultdict(LossyList))
    aspects: Dict[str, Dict[str, int]] = field(
        default_factory=lambda: defaultdict(lambda: defaultdict(int))
    )

    warnings: LossyDict[str, LossyList[str]] = field(default_factory=LossyDict)
    failures: LossyDict[str, LossyList[str]] = field(default_factory=LossyDict)

    def report_workunit(self, wu: WorkUnit) -> None:
        self.events_produced += 1

        if isinstance(wu, MetadataWorkUnit):
            urn = wu.get_urn()

            # Specialized entity reporting.
            if not isinstance(wu.metadata, MetadataChangeEvent):
                mcps = [wu.metadata]
            else:
                mcps = list(mcps_from_mce(wu.metadata))

            for mcp in mcps:
                entityType = mcp.entityType
                aspectName = mcp.aspectName

                if urn not in self._urns_seen:
                    self._urns_seen.add(urn)
                    self.entities[entityType].append(urn)

                if aspectName is not None:  # usually true
                    self.aspects[entityType][aspectName] += 1

    def report_warning(self, key: str, reason: str) -> None:
        warnings = self.warnings.get(key, LossyList())
        warnings.append(reason)
        self.warnings[key] = warnings

    def report_failure(self, key: str, reason: str) -> None:
        failures = self.failures.get(key, LossyList())
        failures.append(reason)
        self.failures[key] = failures

    def __post_init__(self) -> None:
        self.start_time = datetime.datetime.now()
        self.running_time: datetime.timedelta = datetime.timedelta(seconds=0)

    def compute_stats(self) -> None:
        duration = datetime.datetime.now() - self.start_time
        workunits_produced = self.events_produced
        if duration.total_seconds() > 0:
            self.events_produced_per_sec: int = int(
                workunits_produced / duration.total_seconds()
            )
            self.running_time = duration
        else:
            self.read_rate = 0


class CapabilityReport(BaseModel):
    """A report capturing the result of any capability evaluation"""

    capable: bool
    failure_reason: Optional[str] = None
    mitigation_message: Optional[str] = None


@dataclass
class TestConnectionReport(Report):
    internal_failure: Optional[bool] = None
    internal_failure_reason: Optional[str] = None
    basic_connectivity: Optional[CapabilityReport] = None
    capability_report: Optional[
        Dict[Union[SourceCapability, str], CapabilityReport]
    ] = None


WorkUnitType = TypeVar("WorkUnitType", bound=WorkUnit)
ExtractorConfig = TypeVar("ExtractorConfig", bound=ConfigModel)


class Extractor(Generic[WorkUnitType, ExtractorConfig], Closeable, metaclass=ABCMeta):
    ctx: PipelineContext
    config: ExtractorConfig

    @classmethod
    def get_config_class(cls) -> Type[ExtractorConfig]:
        config_class = get_class_from_annotation(cls, Extractor, ConfigModel)
        assert config_class, "Extractor subclasses must define a config class"
        return cast(Type[ExtractorConfig], config_class)

    def __init__(self, config_dict: dict, ctx: PipelineContext) -> None:
        super().__init__()

        config_class = self.get_config_class()

        self.ctx = ctx
        self.config = config_class.parse_obj(config_dict)

    @abstractmethod
    def get_records(self, workunit: WorkUnitType) -> Iterable[RecordEnvelope]:
        pass


@dataclass
class Source(Closeable, metaclass=ABCMeta):
    ctx: PipelineContext

    @classmethod
    def create(cls, config_dict: dict, ctx: PipelineContext) -> "Source":
        # Technically, this method should be abstract. However, the @config_class
        # decorator automatically generates a create method at runtime if one is
        # not defined. Python still treats the class as abstract because it thinks
        # the create method is missing. To avoid the class becoming abstract, we
        # can't make this method abstract.
        raise NotImplementedError('sources must implement "create"')

    @abstractmethod
    def get_workunits(self) -> Iterable[WorkUnit]:
        pass

    @abstractmethod
    def get_report(self) -> SourceReport:
        pass

    def close(self) -> None:
        pass


class TestableSource(Source):
    @staticmethod
    @abstractmethod
    def test_connection(config_dict: dict) -> TestConnectionReport:
        raise NotImplementedError("This class does not implement this method")



from typing import List, Optional, Union

from datahub.emitter.mcp_patch_builder import MetadataPatchProposal
from datahub.metadata.schema_classes import (
    GlossaryTermAssociationClass as Term,
    KafkaAuditHeaderClass,
    OwnerClass as Owner,
    OwnershipTypeClass,
    SystemMetadataClass,
    TagAssociationClass as Tag,
    UpstreamClass as Upstream,
)
from datahub.utilities.urns.tag_urn import TagUrn
from datahub.utilities.urns.urn import Urn


class FieldPatchBuilder(MetadataPatchProposal):
    def __init__(
        self,
        dataset_urn: str,
        field_path: str,
        system_metadata: Optional[SystemMetadataClass] = None,
        audit_header: Optional[KafkaAuditHeaderClass] = None,
        editable: bool = True,
    ) -> None:
        super().__init__(
            dataset_urn,
            "dataset",
            system_metadata=system_metadata,
            audit_header=audit_header,
        )
        self.field_path = field_path
        self.aspect_name = "editableSchemaMetadata" if editable else "schemaMetadata"
        self.aspect_field = "editableSchemaFieldInfo" if editable else "schemaFieldInfo"

    def add_tag(self, tag: Tag) -> "FieldPatchBuilder":
        self._add_patch(
            self.aspect_name,
            "add",
            path=f"/{self.aspect_field}/{self.field_path}/globalTags/tags/{tag.tag}",
            value=tag,
        )
        return self

    def remove_tag(self, tag: Union[str, Urn]) -> "FieldPatchBuilder":
        if isinstance(tag, str) and not tag.startswith("urn:li:tag:"):
            tag = TagUrn.create_from_id(tag)
        self._add_patch(
            self.aspect_name,
            "remove",
            path=f"/{self.aspect_field}/{self.field_path}/globalTags/tags/{tag}",
            value={},
        )
        return self

    def add_term(self, term: Term) -> "FieldPatchBuilder":
        self._add_patch(
            self.aspect_name,
            "add",
            path=f"/{self.aspect_field}/{self.field_path}/glossaryTerms/terms/{term.urn}",
            value=term,
        )
        return self

    def remove_term(self, term: Union[str, Urn]) -> "FieldPatchBuilder":
        if isinstance(term, str) and not term.startswith("urn:li:glossaryTerm:"):
            term = "urn:li:glossaryTerm:" + term
        self._add_patch(
            self.aspect_name,
            "remove",
            path=f"/{self.aspect_field}/{self.field_path}/glossaryTerms/terms/{term}",
            value={},
        )
        return self


class DatasetPatchBuilder(MetadataPatchProposal):
    def __init__(
        self,
        urn: str,
        system_metadata: Optional[SystemMetadataClass] = None,
        audit_header: Optional[KafkaAuditHeaderClass] = None,
    ) -> None:
        super().__init__(
            urn, "dataset", system_metadata=system_metadata, audit_header=audit_header
        )

    def add_owner(self, owner: Owner) -> "DatasetPatchBuilder":
        self._add_patch(
            "ownership", "add", path=f"/owners/{owner.owner}/{owner.type}", value=owner
        )
        return self

    def remove_owner(
        self, owner: Urn, owner_type: Optional[OwnershipTypeClass] = None
    ) -> "DatasetPatchBuilder":
        """
        param: owner_type is optional
        """
        self._add_patch(
            "ownership",
            "remove",
            path=f"/owners/{owner}" + (f"/{owner_type}" if owner_type else ""),
            value=owner,
        )
        return self

    def set_owners(self, owners: List[Owner]) -> "DatasetPatchBuilder":
        self._add_patch("ownership", "replace", path="/owners", value=owners)
        return self

    def add_upstream_lineage(self, upstream: Upstream) -> "DatasetPatchBuilder":
        self._add_patch(
            "upstreamLineage",
            "add",
            path=f"/upstreams/{upstream.dataset}",
            value=upstream,
        )
        return self

    def remove_upstream_lineage(
        self, dataset: Union[str, Urn]
    ) -> "DatasetPatchBuilder":
        self._add_patch(
            "upstreamLineage",
            "remove",
            path=f"/upstreams/{dataset}",
            value={},
        )
        return self

    def set_upstream_lineages(self, upstreams: List[Upstream]) -> "DatasetPatchBuilder":
        self._add_patch(
            "upstreamLineage", "replace", path="/upstreams", value=upstreams
        )
        return self

    def add_tag(self, tag: Tag) -> "DatasetPatchBuilder":
        self._add_patch("globalTags", "add", path=f"/tags/{tag.tag}", value=tag)
        return self

    def remove_tag(self, tag: Union[str, Urn]) -> "DatasetPatchBuilder":
        if isinstance(tag, str) and not tag.startswith("urn:li:tag:"):
            tag = TagUrn.create_from_id(tag)
        self._add_patch("globalTags", "remove", path=f"/tags/{tag}", value={})
        return self

    def add_term(self, term: Term) -> "DatasetPatchBuilder":
        self._add_patch("glossaryTerms", "add", path=f"/terms/{term.urn}", value=term)
        return self

    def remove_term(self, term: Union[str, Urn]) -> "DatasetPatchBuilder":
        if isinstance(term, str) and not term.startswith("urn:li:glossaryTerm:"):
            term = "urn:li:glossaryTerm:" + term
        self._add_patch("glossaryTerms", "remove", path=f"/terms/{term}", value={})
        return self

    def field_patch_builder(
        self, field_path: str, editable: bool = True
    ) -> "FieldPatchBuilder":
        """
        Get a builder that can perform patches against fields in the dataset

        :param field_path: The field path in datahub format
        :param editable: Whether patches should apply to the editable section of schema metadata or not
        """
        return FieldPatchBuilder(
            self.urn,
            field_path,
            system_metadata=self.system_metadata,
            audit_header=self.audit_header,
            editable=editable,
        )



from typing import Optional

from sqllineage.core.holders import SubQueryLineageHolder
from sqllineage.core.models import Column
from sqllineage.exceptions import SQLLineageException
from sqllineage.utils.constant import EdgeType


# Patch based on sqllineage v1.3.3
def end_of_query_cleanup_patch(self, holder: SubQueryLineageHolder) -> None:  # type: ignore
    for i, tbl in enumerate(self.tables):
        holder.add_read(tbl)
    self.union_barriers.append((len(self.columns), len(self.tables)))

    for i, (col_barrier, tbl_barrier) in enumerate(self.union_barriers):
        prev_col_barrier, prev_tbl_barrier = (
            (0, 0) if i == 0 else self.union_barriers[i - 1]
        )
        col_grp = self.columns[prev_col_barrier:col_barrier]
        tbl_grp = self.tables[prev_tbl_barrier:tbl_barrier]
        tgt_tbl = None
        if holder.write:
            if len(holder.write) > 1:
                raise SQLLineageException
            tgt_tbl = list(holder.write)[0]
        elif holder.read:
            tgt_tbl = list(holder.read)[0]

        if tgt_tbl:
            for tgt_col in col_grp:
                tgt_col.parent = tgt_tbl
                for src_col in tgt_col.to_source_columns(
                    self._get_alias_mapping_from_table_group(tbl_grp, holder)
                ):
                    if holder.write:
                        holder.add_column_lineage(src_col, tgt_col)
                    else:
                        holder.add_column_lineage(None, tgt_col)


def add_column_lineage_patch(self, src: Optional[Column], tgt: Column) -> None:  # type: ignore
    if tgt and src:
        self.graph.add_edge(src, tgt, type=EdgeType.LINEAGE)
    if tgt:
        self.graph.add_edge(tgt.parent, tgt, type=EdgeType.HAS_COLUMN)
    if src and src.parent is not None:
        # starting NetworkX v2.6, None is not allowed as node, see https://github.com/networkx/networkx/pull/4892
        self.graph.add_edge(src.parent, src, type=EdgeType.HAS_COLUMN)



import collections
import pathlib
import sqlite3
from dataclasses import dataclass, field
from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    Iterator,
    List,
    MutableMapping,
    Optional,
    OrderedDict,
    Sequence,
    Tuple,
    TypeVar,
    Union,
)

_DEFAULT_TABLE_NAME = "data"
_DEFAULT_MEMORY_CACHE_MAX_SIZE = 2000
_DEFAULT_MEMORY_CACHE_EVICTION_BATCH_SIZE = 200

# https://docs.python.org/3/library/sqlite3.html#sqlite-and-python-types
SqliteValue = Union[int, float, str, bytes, None]

_VT = TypeVar("_VT")


@dataclass
class _SqliteConnectionCache:
    """
    If you pass the same filename to multiple FileBacked* objects, they will
    share the same underlying database connection. This also does ref counting
    to drop the connection when appropriate.
    """

    _ref_count: Dict[pathlib.Path, int] = field(default_factory=dict)
    _sqlite_connection_cache: Dict[pathlib.Path, sqlite3.Connection] = field(
        default_factory=dict
    )

    def get_connection(self, filename: pathlib.Path) -> sqlite3.Connection:
        if filename not in self._ref_count:
            conn = sqlite3.connect(filename, isolation_level=None)

            # These settings are optimized for performance.
            # See https://www.sqlite.org/pragma.html for more information.
            # Because we're only using these dbs to offload data from memory, we don't need
            # to worry about data integrity too much.
            conn.execute('PRAGMA locking_mode = "EXCLUSIVE"')
            conn.execute('PRAGMA synchronous = "OFF"')
            conn.execute('PRAGMA journal_mode = "MEMORY"')
            conn.execute(f"PRAGMA journal_size_limit = {100 * 1024 * 1024}")  # 100MB

            self._ref_count[filename] = 0
            self._sqlite_connection_cache[filename] = conn

        self._ref_count[filename] += 1
        return self._sqlite_connection_cache[filename]

    def drop_connection(self, filename: pathlib.Path) -> None:
        self._ref_count[filename] -= 1

        if self._ref_count[filename] == 0:
            # Cleanup the connection object.
            self._sqlite_connection_cache[filename].close()
            del self._sqlite_connection_cache[filename]
            del self._ref_count[filename]


_sqlite_connection_cache = _SqliteConnectionCache()


@dataclass(eq=False)
class FileBackedDict(MutableMapping[str, _VT], Generic[_VT]):
    """
    A dict-like object that stores its data in a temporary SQLite database.
    This is useful for storing large amounts of data that don't fit in memory.

    This class is not thread-safe.
    """

    filename: pathlib.Path

    serializer: Callable[[_VT], SqliteValue]
    deserializer: Callable[[Any], _VT]

    tablename: str = field(default=_DEFAULT_TABLE_NAME)
    extra_columns: Dict[str, Callable[[_VT], SqliteValue]] = field(default_factory=dict)

    cache_max_size: int = field(default=_DEFAULT_MEMORY_CACHE_MAX_SIZE)
    cache_eviction_batch_size: int = field(
        default=_DEFAULT_MEMORY_CACHE_EVICTION_BATCH_SIZE
    )

    _conn: sqlite3.Connection = field(init=False, repr=False)

    # To improve performance, we maintain an in-memory LRU cache using an OrderedDict.
    _active_object_cache: OrderedDict[str, _VT] = field(init=False, repr=False)

    def __post_init__(self) -> None:
        assert (
            self.cache_eviction_batch_size > 0
        ), "cache_eviction_batch_size must be positive"

        self._conn = _sqlite_connection_cache.get_connection(self.filename)

        # We keep a small cache in memory to avoid having to serialize/deserialize
        # data from the database too often. We use an OrderedDict to build
        # a poor-man's LRU cache.
        self._active_object_cache = collections.OrderedDict()

        # Create the table. We're not using "IF NOT EXISTS" because creating
        # the same table twice indicates a client usage error.
        self._conn.execute(
            f"""CREATE TABLE {self.tablename} (
                key TEXT PRIMARY KEY,
                value BLOB
                {''.join(f', {column_name} BLOB' for column_name in self.extra_columns.keys())}
            )"""
        )

        # The key column will automatically be indexed, but we need indexes
        # for the extra columns.
        for column_name in self.extra_columns.keys():
            self._conn.execute(
                f"CREATE INDEX {self.tablename}_{column_name} ON {self.tablename} ({column_name})"
            )

    def _add_to_cache(self, key: str, value: _VT) -> None:
        self._active_object_cache[key] = value

        if len(self._active_object_cache) > self.cache_max_size:
            # Try to prune in batches rather than one at a time.
            num_items_to_prune = min(
                len(self._active_object_cache), self.cache_eviction_batch_size
            )
            self._prune_cache(num_items_to_prune)

    def _prune_cache(self, num_items_to_prune: int) -> None:
        items_to_write: List[Tuple[SqliteValue, ...]] = []
        for _ in range(num_items_to_prune):
            key, value = self._active_object_cache.popitem(last=False)

            values = [key, self.serializer(value)]
            for column_serializer in self.extra_columns.values():
                values.append(column_serializer(value))
            items_to_write.append(tuple(values))

        self._conn.executemany(
            f"""INSERT OR REPLACE INTO {self.tablename} (
                key,
                value
                {''.join(f', {column_name}' for column_name in self.extra_columns.keys())}
            )
            VALUES ({', '.join(['?'] *(2 + len(self.extra_columns)))})""",
            items_to_write,
        )

    def flush(self) -> None:
        self._prune_cache(len(self._active_object_cache))

    def __getitem__(self, key: str) -> _VT:
        if key in self._active_object_cache:
            self._active_object_cache.move_to_end(key)
            return self._active_object_cache[key]

        cursor = self._conn.execute(
            f"SELECT value FROM {self.tablename} WHERE key = ?", (key,)
        )
        result: Sequence[SqliteValue] = cursor.fetchone()
        if result is None:
            raise KeyError(key)

        deserialized_result = self.deserializer(result[0])
        self._add_to_cache(key, deserialized_result)
        return deserialized_result

    def __setitem__(self, key: str, value: _VT) -> None:
        self._add_to_cache(key, value)

    def __delitem__(self, key: str) -> None:
        in_cache = False
        if key in self._active_object_cache:
            del self._active_object_cache[key]
            in_cache = True

        n_deleted = self._conn.execute(
            f"DELETE FROM {self.tablename} WHERE key = ?", (key,)
        ).rowcount
        if not in_cache and not n_deleted:
            raise KeyError(key)

    def __iter__(self) -> Iterator[str]:
        cursor = self._conn.execute(f"SELECT key FROM {self.tablename}")
        for row in cursor:
            if row[0] in self._active_object_cache:
                # If the key is in the active object cache, then SQL isn't the source of truth.
                continue

            yield row[0]

        for key in self._active_object_cache:
            yield key

    def __len__(self) -> int:
        cursor = self._conn.execute(
            # Binding a list of values in SQLite: https://stackoverflow.com/a/1310001/5004662.
            f"SELECT COUNT(*) FROM {self.tablename} WHERE key NOT IN ({','.join('?' * len(self._active_object_cache))})",
            (*self._active_object_cache.keys(),),
        )
        row = cursor.fetchone()

        return row[0] + len(self._active_object_cache)

    def sql_query(
        self,
        query: str,
        params: Tuple[Any, ...] = (),
        refs: Optional[List[Union["FileBackedList", "FileBackedDict"]]] = None,
    ) -> List[Tuple[Any, ...]]:
        # We need to flush object and any objects the query references to ensure
        # that we don't miss objects that have been modified but not yet flushed.
        self.flush()
        if refs is not None:
            for referenced_table in refs:
                referenced_table.flush()

        cursor = self._conn.execute(query, params)
        return cursor.fetchall()

    def close(self) -> None:
        _sqlite_connection_cache.drop_connection(self.filename)

    def __del__(self) -> None:
        self.close()


class FileBackedList(Generic[_VT]):
    """
    An append-only, list-like object that stores its contents in a SQLite database.

    This class is not thread-safe.
    """

    _len: int = field(default=0)
    _dict: FileBackedDict[_VT] = field(init=False)

    def __init__(
        self,
        filename: pathlib.Path,
        serializer: Callable[[_VT], SqliteValue],
        deserializer: Callable[[Any], _VT],
        tablename: str = _DEFAULT_TABLE_NAME,
        extra_columns: Optional[Dict[str, Callable[[_VT], SqliteValue]]] = None,
        cache_max_size: Optional[int] = None,
        cache_eviction_batch_size: Optional[int] = None,
    ) -> None:
        self._len = 0
        self._dict = FileBackedDict(
            filename=filename,
            serializer=serializer,
            deserializer=deserializer,
            tablename=tablename,
            extra_columns=extra_columns or {},
            cache_max_size=cache_max_size or _DEFAULT_MEMORY_CACHE_MAX_SIZE,
            cache_eviction_batch_size=cache_eviction_batch_size
            or _DEFAULT_MEMORY_CACHE_EVICTION_BATCH_SIZE,
        )

    @property
    def tablename(self) -> str:
        return self._dict.tablename

    def __getitem__(self, index: int) -> _VT:
        if index < 0 or index >= self._len:
            raise IndexError(f"list index {index} out of range")

        return self._dict[str(index)]

    def __setitem__(self, index: int, value: _VT) -> None:
        if index < 0 or index >= self._len:
            raise IndexError(f"list index {index} out of range")

        self._dict[str(index)] = value

    def append(self, value: _VT) -> None:
        self._dict[str(self._len)] = value
        self._len += 1

    def __len__(self) -> int:
        return self._len

    def __iter__(self) -> Iterator[_VT]:
        for index in range(self._len):
            yield self[index]

    def flush(self) -> None:
        self._dict.flush()

    def sql_query(
        self,
        query: str,
        params: Tuple[Any, ...] = (),
        refs: Optional[List[Union["FileBackedList", "FileBackedDict"]]] = None,
    ) -> List[Tuple[Any, ...]]:
        return self._dict.sql_query(query, params, refs=refs)



# flake8: noqa

# This file is autogenerated by /metadata-ingestion/scripts/avro_codegen.py
# Do not modify manually!

# pylint: skip-file
# fmt: off

import functools
import pathlib

def _load_schema(schema_name: str) -> str:
    return (pathlib.Path(__file__).parent / f"{schema_name}.avsc").read_text()

@functools.lru_cache(maxsize=None)
def getTelemetryClientIdSchema() -> str:
    return _load_schema("TelemetryClientId")

@functools.lru_cache(maxsize=None)
def getDataProcessInstanceRelationshipsSchema() -> str:
    return _load_schema("DataProcessInstanceRelationships")

@functools.lru_cache(maxsize=None)
def getDataProcessInstancePropertiesSchema() -> str:
    return _load_schema("DataProcessInstanceProperties")

@functools.lru_cache(maxsize=None)
def getDataProcessInstanceOutputSchema() -> str:
    return _load_schema("DataProcessInstanceOutput")

@functools.lru_cache(maxsize=None)
def getDataProcessInstanceInputSchema() -> str:
    return _load_schema("DataProcessInstanceInput")

@functools.lru_cache(maxsize=None)
def getDataProcessInfoSchema() -> str:
    return _load_schema("DataProcessInfo")

@functools.lru_cache(maxsize=None)
def getDataProcessInstanceRunEventSchema() -> str:
    return _load_schema("DataProcessInstanceRunEvent")

@functools.lru_cache(maxsize=None)
def getAssertionInfoSchema() -> str:
    return _load_schema("AssertionInfo")

@functools.lru_cache(maxsize=None)
def getAssertionRunEventSchema() -> str:
    return _load_schema("AssertionRunEvent")

@functools.lru_cache(maxsize=None)
def getGlossaryTermInfoSchema() -> str:
    return _load_schema("GlossaryTermInfo")

@functools.lru_cache(maxsize=None)
def getGlossaryRelatedTermsSchema() -> str:
    return _load_schema("GlossaryRelatedTerms")

@functools.lru_cache(maxsize=None)
def getGlossaryNodeInfoSchema() -> str:
    return _load_schema("GlossaryNodeInfo")

@functools.lru_cache(maxsize=None)
def getExecutionRequestResultSchema() -> str:
    return _load_schema("ExecutionRequestResult")

@functools.lru_cache(maxsize=None)
def getExecutionRequestSignalSchema() -> str:
    return _load_schema("ExecutionRequestSignal")

@functools.lru_cache(maxsize=None)
def getExecutionRequestInputSchema() -> str:
    return _load_schema("ExecutionRequestInput")

@functools.lru_cache(maxsize=None)
def getMetadataChangeProposalSchema() -> str:
    return _load_schema("MetadataChangeProposal")

@functools.lru_cache(maxsize=None)
def getMetadataChangeEventSchema() -> str:
    return _load_schema("MetadataChangeEvent")

@functools.lru_cache(maxsize=None)
def getMetadataChangeLogSchema() -> str:
    return _load_schema("MetadataChangeLog")

@functools.lru_cache(maxsize=None)
def getPlatformEventSchema() -> str:
    return _load_schema("PlatformEvent")

@functools.lru_cache(maxsize=None)
def getDatasetProfileSchema() -> str:
    return _load_schema("DatasetProfile")

@functools.lru_cache(maxsize=None)
def getViewPropertiesSchema() -> str:
    return _load_schema("ViewProperties")

@functools.lru_cache(maxsize=None)
def getEditableDatasetPropertiesSchema() -> str:
    return _load_schema("EditableDatasetProperties")

@functools.lru_cache(maxsize=None)
def getDatasetUpstreamLineageSchema() -> str:
    return _load_schema("DatasetUpstreamLineage")

@functools.lru_cache(maxsize=None)
def getDatasetDeprecationSchema() -> str:
    return _load_schema("DatasetDeprecation")

@functools.lru_cache(maxsize=None)
def getDatasetPropertiesSchema() -> str:
    return _load_schema("DatasetProperties")

@functools.lru_cache(maxsize=None)
def getDatasetUsageStatisticsSchema() -> str:
    return _load_schema("DatasetUsageStatistics")

@functools.lru_cache(maxsize=None)
def getUpstreamLineageSchema() -> str:
    return _load_schema("UpstreamLineage")

@functools.lru_cache(maxsize=None)
def getEditableDashboardPropertiesSchema() -> str:
    return _load_schema("EditableDashboardProperties")

@functools.lru_cache(maxsize=None)
def getDashboardInfoSchema() -> str:
    return _load_schema("DashboardInfo")

@functools.lru_cache(maxsize=None)
def getDashboardUsageStatisticsSchema() -> str:
    return _load_schema("DashboardUsageStatistics")

@functools.lru_cache(maxsize=None)
def getTestInfoSchema() -> str:
    return _load_schema("TestInfo")

@functools.lru_cache(maxsize=None)
def getTestResultsSchema() -> str:
    return _load_schema("TestResults")

@functools.lru_cache(maxsize=None)
def getMLHyperParamSchema() -> str:
    return _load_schema("MLHyperParam")

@functools.lru_cache(maxsize=None)
def getIntendedUseSchema() -> str:
    return _load_schema("IntendedUse")

@functools.lru_cache(maxsize=None)
def getMLModelDeploymentPropertiesSchema() -> str:
    return _load_schema("MLModelDeploymentProperties")

@functools.lru_cache(maxsize=None)
def getMLModelGroupPropertiesSchema() -> str:
    return _load_schema("MLModelGroupProperties")

@functools.lru_cache(maxsize=None)
def getMLMetricSchema() -> str:
    return _load_schema("MLMetric")

@functools.lru_cache(maxsize=None)
def getQuantitativeAnalysesSchema() -> str:
    return _load_schema("QuantitativeAnalyses")

@functools.lru_cache(maxsize=None)
def getEvaluationDataSchema() -> str:
    return _load_schema("EvaluationData")

@functools.lru_cache(maxsize=None)
def getEthicalConsiderationsSchema() -> str:
    return _load_schema("EthicalConsiderations")

@functools.lru_cache(maxsize=None)
def getEditableMLFeatureTablePropertiesSchema() -> str:
    return _load_schema("EditableMLFeatureTableProperties")

@functools.lru_cache(maxsize=None)
def getEditableMLModelGroupPropertiesSchema() -> str:
    return _load_schema("EditableMLModelGroupProperties")

@functools.lru_cache(maxsize=None)
def getMLFeatureTablePropertiesSchema() -> str:
    return _load_schema("MLFeatureTableProperties")

@functools.lru_cache(maxsize=None)
def getEditableMLPrimaryKeyPropertiesSchema() -> str:
    return _load_schema("EditableMLPrimaryKeyProperties")

@functools.lru_cache(maxsize=None)
def getEditableMLModelPropertiesSchema() -> str:
    return _load_schema("EditableMLModelProperties")

@functools.lru_cache(maxsize=None)
def getMLModelFactorPromptsSchema() -> str:
    return _load_schema("MLModelFactorPrompts")

@functools.lru_cache(maxsize=None)
def getMetricsSchema() -> str:
    return _load_schema("Metrics")

@functools.lru_cache(maxsize=None)
def getEditableMLFeaturePropertiesSchema() -> str:
    return _load_schema("EditableMLFeatureProperties")

@functools.lru_cache(maxsize=None)
def getMLFeaturePropertiesSchema() -> str:
    return _load_schema("MLFeatureProperties")

@functools.lru_cache(maxsize=None)
def getCaveatsAndRecommendationsSchema() -> str:
    return _load_schema("CaveatsAndRecommendations")

@functools.lru_cache(maxsize=None)
def getTrainingDataSchema() -> str:
    return _load_schema("TrainingData")

@functools.lru_cache(maxsize=None)
def getSourceCodeSchema() -> str:
    return _load_schema("SourceCode")

@functools.lru_cache(maxsize=None)
def getMLPrimaryKeyPropertiesSchema() -> str:
    return _load_schema("MLPrimaryKeyProperties")

@functools.lru_cache(maxsize=None)
def getMLModelPropertiesSchema() -> str:
    return _load_schema("MLModelProperties")

@functools.lru_cache(maxsize=None)
def getGlossaryTermsSchema() -> str:
    return _load_schema("GlossaryTerms")

@functools.lru_cache(maxsize=None)
def getStatusSchema() -> str:
    return _load_schema("Status")

@functools.lru_cache(maxsize=None)
def getInputFieldsSchema() -> str:
    return _load_schema("InputFields")

@functools.lru_cache(maxsize=None)
def getCostSchema() -> str:
    return _load_schema("Cost")

@functools.lru_cache(maxsize=None)
def getDataPlatformInstanceSchema() -> str:
    return _load_schema("DataPlatformInstance")

@functools.lru_cache(maxsize=None)
def getBrowsePathsSchema() -> str:
    return _load_schema("BrowsePaths")

@functools.lru_cache(maxsize=None)
def getOriginSchema() -> str:
    return _load_schema("Origin")

@functools.lru_cache(maxsize=None)
def getOperationSchema() -> str:
    return _load_schema("Operation")

@functools.lru_cache(maxsize=None)
def getSubTypesSchema() -> str:
    return _load_schema("SubTypes")

@functools.lru_cache(maxsize=None)
def getOwnershipSchema() -> str:
    return _load_schema("Ownership")

@functools.lru_cache(maxsize=None)
def getEmbedSchema() -> str:
    return _load_schema("Embed")

@functools.lru_cache(maxsize=None)
def getSiblingsSchema() -> str:
    return _load_schema("Siblings")

@functools.lru_cache(maxsize=None)
def getInstitutionalMemorySchema() -> str:
    return _load_schema("InstitutionalMemory")

@functools.lru_cache(maxsize=None)
def getDeprecationSchema() -> str:
    return _load_schema("Deprecation")

@functools.lru_cache(maxsize=None)
def getGlobalTagsSchema() -> str:
    return _load_schema("GlobalTags")

@functools.lru_cache(maxsize=None)
def getDataPlatformInfoSchema() -> str:
    return _load_schema("DataPlatformInfo")

@functools.lru_cache(maxsize=None)
def getEntityChangeEventSchema() -> str:
    return _load_schema("EntityChangeEvent")

@functools.lru_cache(maxsize=None)
def getEditableContainerPropertiesSchema() -> str:
    return _load_schema("EditableContainerProperties")

@functools.lru_cache(maxsize=None)
def getContainerSchema() -> str:
    return _load_schema("Container")

@functools.lru_cache(maxsize=None)
def getContainerPropertiesSchema() -> str:
    return _load_schema("ContainerProperties")

@functools.lru_cache(maxsize=None)
def getGlobalSettingsInfoSchema() -> str:
    return _load_schema("GlobalSettingsInfo")

@functools.lru_cache(maxsize=None)
def getDataHubRetentionConfigSchema() -> str:
    return _load_schema("DataHubRetentionConfig")

@functools.lru_cache(maxsize=None)
def getDomainsSchema() -> str:
    return _load_schema("Domains")

@functools.lru_cache(maxsize=None)
def getDomainPropertiesSchema() -> str:
    return _load_schema("DomainProperties")

@functools.lru_cache(maxsize=None)
def getDataHubIngestionSourceInfoSchema() -> str:
    return _load_schema("DataHubIngestionSourceInfo")

@functools.lru_cache(maxsize=None)
def getEditableSchemaMetadataSchema() -> str:
    return _load_schema("EditableSchemaMetadata")

@functools.lru_cache(maxsize=None)
def getSchemaMetadataSchema() -> str:
    return _load_schema("SchemaMetadata")

@functools.lru_cache(maxsize=None)
def getFilterSchema() -> str:
    return _load_schema("Filter")

@functools.lru_cache(maxsize=None)
def getGlobalSettingsKeySchema() -> str:
    return _load_schema("GlobalSettingsKey")

@functools.lru_cache(maxsize=None)
def getPostKeySchema() -> str:
    return _load_schema("PostKey")

@functools.lru_cache(maxsize=None)
def getTagKeySchema() -> str:
    return _load_schema("TagKey")

@functools.lru_cache(maxsize=None)
def getDataPlatformInstanceKeySchema() -> str:
    return _load_schema("DataPlatformInstanceKey")

@functools.lru_cache(maxsize=None)
def getDataHubSecretKeySchema() -> str:
    return _load_schema("DataHubSecretKey")

@functools.lru_cache(maxsize=None)
def getMLFeatureKeySchema() -> str:
    return _load_schema("MLFeatureKey")

@functools.lru_cache(maxsize=None)
def getDataJobKeySchema() -> str:
    return _load_schema("DataJobKey")

@functools.lru_cache(maxsize=None)
def getGlossaryNodeKeySchema() -> str:
    return _load_schema("GlossaryNodeKey")

@functools.lru_cache(maxsize=None)
def getDataProcessKeySchema() -> str:
    return _load_schema("DataProcessKey")

@functools.lru_cache(maxsize=None)
def getCorpUserKeySchema() -> str:
    return _load_schema("CorpUserKey")

@functools.lru_cache(maxsize=None)
def getDataHubViewKeySchema() -> str:
    return _load_schema("DataHubViewKey")

@functools.lru_cache(maxsize=None)
def getDataHubRoleKeySchema() -> str:
    return _load_schema("DataHubRoleKey")

@functools.lru_cache(maxsize=None)
def getDataHubPolicyKeySchema() -> str:
    return _load_schema("DataHubPolicyKey")

@functools.lru_cache(maxsize=None)
def getMLModelKeySchema() -> str:
    return _load_schema("MLModelKey")

@functools.lru_cache(maxsize=None)
def getMLFeatureTableKeySchema() -> str:
    return _load_schema("MLFeatureTableKey")

@functools.lru_cache(maxsize=None)
def getTelemetryKeySchema() -> str:
    return _load_schema("TelemetryKey")

@functools.lru_cache(maxsize=None)
def getChartKeySchema() -> str:
    return _load_schema("ChartKey")

@functools.lru_cache(maxsize=None)
def getQueryKeySchema() -> str:
    return _load_schema("QueryKey")

@functools.lru_cache(maxsize=None)
def getDataHubUpgradeKeySchema() -> str:
    return _load_schema("DataHubUpgradeKey")

@functools.lru_cache(maxsize=None)
def getDataHubRetentionKeySchema() -> str:
    return _load_schema("DataHubRetentionKey")

@functools.lru_cache(maxsize=None)
def getDatasetKeySchema() -> str:
    return _load_schema("DatasetKey")

@functools.lru_cache(maxsize=None)
def getTestKeySchema() -> str:
    return _load_schema("TestKey")

@functools.lru_cache(maxsize=None)
def getMLModelGroupKeySchema() -> str:
    return _load_schema("MLModelGroupKey")

@functools.lru_cache(maxsize=None)
def getDataHubIngestionSourceKeySchema() -> str:
    return _load_schema("DataHubIngestionSourceKey")

@functools.lru_cache(maxsize=None)
def getMLModelDeploymentKeySchema() -> str:
    return _load_schema("MLModelDeploymentKey")

@functools.lru_cache(maxsize=None)
def getInviteTokenKeySchema() -> str:
    return _load_schema("InviteTokenKey")

@functools.lru_cache(maxsize=None)
def getDomainKeySchema() -> str:
    return _load_schema("DomainKey")

@functools.lru_cache(maxsize=None)
def getDataPlatformKeySchema() -> str:
    return _load_schema("DataPlatformKey")

@functools.lru_cache(maxsize=None)
def getContainerKeySchema() -> str:
    return _load_schema("ContainerKey")

@functools.lru_cache(maxsize=None)
def getDataFlowKeySchema() -> str:
    return _load_schema("DataFlowKey")

@functools.lru_cache(maxsize=None)
def getAssertionKeySchema() -> str:
    return _load_schema("AssertionKey")

@functools.lru_cache(maxsize=None)
def getGlossaryTermKeySchema() -> str:
    return _load_schema("GlossaryTermKey")

@functools.lru_cache(maxsize=None)
def getExecutionRequestKeySchema() -> str:
    return _load_schema("ExecutionRequestKey")

@functools.lru_cache(maxsize=None)
def getSchemaFieldKeySchema() -> str:
    return _load_schema("SchemaFieldKey")

@functools.lru_cache(maxsize=None)
def getDataHubStepStateKeySchema() -> str:
    return _load_schema("DataHubStepStateKey")

@functools.lru_cache(maxsize=None)
def getMLPrimaryKeyKeySchema() -> str:
    return _load_schema("MLPrimaryKeyKey")

@functools.lru_cache(maxsize=None)
def getCorpGroupKeySchema() -> str:
    return _load_schema("CorpGroupKey")

@functools.lru_cache(maxsize=None)
def getNotebookKeySchema() -> str:
    return _load_schema("NotebookKey")

@functools.lru_cache(maxsize=None)
def getDashboardKeySchema() -> str:
    return _load_schema("DashboardKey")

@functools.lru_cache(maxsize=None)
def getDataHubAccessTokenKeySchema() -> str:
    return _load_schema("DataHubAccessTokenKey")

@functools.lru_cache(maxsize=None)
def getDataProcessInstanceKeySchema() -> str:
    return _load_schema("DataProcessInstanceKey")

@functools.lru_cache(maxsize=None)
def getUsageAggregationSchema() -> str:
    return _load_schema("UsageAggregation")

@functools.lru_cache(maxsize=None)
def getDataHubStepStatePropertiesSchema() -> str:
    return _load_schema("DataHubStepStateProperties")

@functools.lru_cache(maxsize=None)
def getTagPropertiesSchema() -> str:
    return _load_schema("TagProperties")

@functools.lru_cache(maxsize=None)
def getDataHubAccessTokenInfoSchema() -> str:
    return _load_schema("DataHubAccessTokenInfo")

@functools.lru_cache(maxsize=None)
def getEditableDataFlowPropertiesSchema() -> str:
    return _load_schema("EditableDataFlowProperties")

@functools.lru_cache(maxsize=None)
def getEditableDataJobPropertiesSchema() -> str:
    return _load_schema("EditableDataJobProperties")

@functools.lru_cache(maxsize=None)
def getVersionInfoSchema() -> str:
    return _load_schema("VersionInfo")

@functools.lru_cache(maxsize=None)
def getDataJobInputOutputSchema() -> str:
    return _load_schema("DataJobInputOutput")

@functools.lru_cache(maxsize=None)
def getDataJobInfoSchema() -> str:
    return _load_schema("DataJobInfo")

@functools.lru_cache(maxsize=None)
def getDataFlowInfoSchema() -> str:
    return _load_schema("DataFlowInfo")

@functools.lru_cache(maxsize=None)
def getDatahubIngestionRunSummarySchema() -> str:
    return _load_schema("DatahubIngestionRunSummary")

@functools.lru_cache(maxsize=None)
def getDatahubIngestionCheckpointSchema() -> str:
    return _load_schema("DatahubIngestionCheckpoint")

@functools.lru_cache(maxsize=None)
def getDataHubSecretValueSchema() -> str:
    return _load_schema("DataHubSecretValue")

@functools.lru_cache(maxsize=None)
def getCorpUserSettingsSchema() -> str:
    return _load_schema("CorpUserSettings")

@functools.lru_cache(maxsize=None)
def getCorpUserStatusSchema() -> str:
    return _load_schema("CorpUserStatus")

@functools.lru_cache(maxsize=None)
def getRoleMembershipSchema() -> str:
    return _load_schema("RoleMembership")

@functools.lru_cache(maxsize=None)
def getCorpUserCredentialsSchema() -> str:
    return _load_schema("CorpUserCredentials")

@functools.lru_cache(maxsize=None)
def getCorpGroupInfoSchema() -> str:
    return _load_schema("CorpGroupInfo")

@functools.lru_cache(maxsize=None)
def getInviteTokenSchema() -> str:
    return _load_schema("InviteToken")

@functools.lru_cache(maxsize=None)
def getGroupMembershipSchema() -> str:
    return _load_schema("GroupMembership")

@functools.lru_cache(maxsize=None)
def getCorpGroupEditableInfoSchema() -> str:
    return _load_schema("CorpGroupEditableInfo")

@functools.lru_cache(maxsize=None)
def getCorpUserEditableInfoSchema() -> str:
    return _load_schema("CorpUserEditableInfo")

@functools.lru_cache(maxsize=None)
def getNativeGroupMembershipSchema() -> str:
    return _load_schema("NativeGroupMembership")

@functools.lru_cache(maxsize=None)
def getCorpUserInfoSchema() -> str:
    return _load_schema("CorpUserInfo")

@functools.lru_cache(maxsize=None)
def getDataHubViewInfoSchema() -> str:
    return _load_schema("DataHubViewInfo")

@functools.lru_cache(maxsize=None)
def getChartInfoSchema() -> str:
    return _load_schema("ChartInfo")

@functools.lru_cache(maxsize=None)
def getChartQuerySchema() -> str:
    return _load_schema("ChartQuery")

@functools.lru_cache(maxsize=None)
def getEditableChartPropertiesSchema() -> str:
    return _load_schema("EditableChartProperties")

@functools.lru_cache(maxsize=None)
def getChartUsageStatisticsSchema() -> str:
    return _load_schema("ChartUsageStatistics")

@functools.lru_cache(maxsize=None)
def getQueryPropertiesSchema() -> str:
    return _load_schema("QueryProperties")

@functools.lru_cache(maxsize=None)
def getQuerySubjectsSchema() -> str:
    return _load_schema("QuerySubjects")

@functools.lru_cache(maxsize=None)
def getDataHubRoleInfoSchema() -> str:
    return _load_schema("DataHubRoleInfo")

@functools.lru_cache(maxsize=None)
def getDataHubPolicyInfoSchema() -> str:
    return _load_schema("DataHubPolicyInfo")

@functools.lru_cache(maxsize=None)
def getEditableNotebookPropertiesSchema() -> str:
    return _load_schema("EditableNotebookProperties")

@functools.lru_cache(maxsize=None)
def getNotebookInfoSchema() -> str:
    return _load_schema("NotebookInfo")

@functools.lru_cache(maxsize=None)
def getNotebookContentSchema() -> str:
    return _load_schema("NotebookContent")

@functools.lru_cache(maxsize=None)
def getDataPlatformInstancePropertiesSchema() -> str:
    return _load_schema("DataPlatformInstanceProperties")

@functools.lru_cache(maxsize=None)
def getPostInfoSchema() -> str:
    return _load_schema("PostInfo")

@functools.lru_cache(maxsize=None)
def getDataHubUpgradeResultSchema() -> str:
    return _load_schema("DataHubUpgradeResult")

@functools.lru_cache(maxsize=None)
def getDataHubUpgradeRequestSchema() -> str:
    return _load_schema("DataHubUpgradeRequest")

# fmt: on
